import asyncio
import tempfile
from dataclasses import dataclass
import logging
from typing import List, Tuple
import os
from subprocess import CalledProcessError

from ofrak.model.tag_model import ResourceTag

from ofrak import Identifier, Analyzer
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.model.component_model import ComponentExternalTool
from ofrak.resource import Resource
from ofrak.core.filesystem import File
from ofrak.core.binary import GenericBinary
from ofrak.resource_view import ResourceView

from ofrak_type.range import Range

try:
    from ubireader import ubi_io
    from ubireader.ubi import ubi as ubireader_ubi
    from ubireader.ubi.defines import UBI_EC_HDR_MAGIC, UBI_VID_HDR_MAGIC, PRINT_VOL_TYPE_LIST
    from ubireader.utils import guess_peb_size
except ModuleNotFoundError:
    pass

from ofrak.model.resource_model import index

LOGGER = logging.getLogger(__name__)

UBINIZE_TOOL = ComponentExternalTool(
    "ubinize",
    "http://www.linux-mtd.infradead.org/faq/ubi.html#L_ubi_mkimg",
    install_check_arg="--help",
    apt_package="mtd-utils",
    brew_package="",  # This isn't compatible with macos, but there may be an alternative tool to do the bidding.
)


class _PyLzoTool(ComponentExternalTool):
    def __init__(self):
        super().__init__(
            "python-lzo",
            "https://github.com/jd-boyd/python-lzo",
            install_check_arg="",
        )

    async def is_tool_installed(self) -> bool:
        try:
            import lzo  # type: ignore

            return True
        except ModuleNotFoundError:
            return False


PY_LZO_TOOL = _PyLzoTool()


@dataclass
class UbiVolume(ResourceView):
    """
    An UbiVolume is a volume entry in UBI. It contains an image of arbitrary data, typically a filesystem or a
    log. Empty UbiVolumes can still reserve physical erase blocks, in case they are expected to grow.

    Volume information reflected in the 'config.ini' UBI volume entries expected by `ubinize` are stored here. Also see:
    http://www.linux-mtd.infradead.org/faq/ubi.html#L_ubi_mkimg

    :var id: The assigned volume ID within the UBI image
    :var peb_count: Number of PEBs allocated for the volume
    :var type: UBI volume type, either 'static' or 'dynamic'; see http://www.linux-mtd.infradead.org/doc/ubi.html#L_overview
    :var name: Label assigned to the volume
    :var flag_autoresize: Tells UBI to resize this volume to occupy unused space in the UBI image once; see http://www.linux-mtd.infradead.org/doc/ubi.html#L_autoresize
    :var alignment: LEB size of this volume has to be aligned on, such that Ubi.leb_size % UbiVolume.alignment == 0; see https://elixir.bootlin.com/linux/v6.1.7/source/include/uapi/mtd/ubi-user.h#L328
    """

    id: int
    peb_count: int  # size = UbiVolume.peb_count * Ubi.peb_size
    type: str
    name: str
    flag_autoresize: bool  # `vol_flags` only specifies an auto-resize flag
    alignment: int

    @index
    def UbiVolumeId(self) -> int:
        return self.id


@dataclass
class Ubi(GenericBinary):
    """
    UBI is a volume management layer for raw / unmanaged flash devices. It can be thought of LVM
    (Logical Volume Manager) with some additional features necessary for reliably using raw flash memory such as wear
    leveling and error correction. Each volume can contain any arbitrary data, though UBIFS is specially made to be used
    within an UBI volume.

    UBI parameters and volumes required by `ubinize` for repacking are defined here. Also see:
    http://www.linux-mtd.infradead.org/doc/ubi.html and
    https://github.com/vamanea/mtd-utils/blob/master/ubi-utils/ubinize.c#L288`

    :var min_io_size: Minimum number of bytes per I/O transaction (see http://www.linux-mtd.infradead.org/doc/ubi.html#L_min_io_unit)
    :var leb_size: Size of Logical Erase Blocks
    :var peb_size: Size of Physical Erase Blocks
    :var total_peb_count: The total number of PEBs, which includes hidden layout blocks in addition to data blocks allocated per volume
    :var image_seq: image sequence number recorded on EC headers (typically random)
    :var volumes: List of volumes associated with the UBI image
    """

    min_io_size: int
    leb_size: int
    peb_size: int
    total_peb_count: int  # total_peb_count usually adds 2 layout blocks in addition to data blocks allocated per volume

    # Each UBI image emitted by `ubireader` would be treated as a separate Ubi resource.
    image_seq: int
    volumes: List[UbiVolume]


class UbiAnalyzer(Analyzer[None, Ubi]):
    """
    Extract UBI parameters required for packing a resource.
    """

    targets = (Ubi,)
    outputs = (Ubi,)

    external_dependencies = (PY_LZO_TOOL,)

    async def analyze(self, resource: Resource, config=None) -> Ubi:
        # Flush to disk
        with tempfile.NamedTemporaryFile() as temp_file:
            resource_data = await resource.get_data()
            temp_file.write(resource_data)
            temp_file.flush()

            ubi_obj = ubireader_ubi(
                ubi_io.ubi_file(
                    temp_file.name,
                    block_size=guess_peb_size(temp_file.name),
                    start_offset=0,
                    end_offset=None,
                )
            )

        # Technically multiple images can be encountered in an UBI blob, but that should be handled by
        # OFRAK by treating them as separate UBI resources.
        if len(ubi_obj.images) > 1:
            raise Exception(
                "Multi-image UBI blobs are not directly supported. Carve each image into a separate "
                "resource and run UbiAnalyzer on each of them."
            )
        if len(ubi_obj.images) == 0:
            raise Exception("UBI resource does not have any images.")

        image = ubi_obj.images[0]
        ubi_image_vols: List[UbiVolume] = []
        for volume in image.volumes.values():
            ubi_image_vols.append(
                UbiVolume(
                    volume.vol_rec.rec_index,
                    volume.vol_rec.reserved_pebs,
                    PRINT_VOL_TYPE_LIST[volume.vol_rec.vol_type],
                    volume.vol_rec.name.decode("utf-8"),
                    volume.vol_rec.flags,  # Autoresize flag for standard UBI
                    volume.vol_rec.alignment,
                )
            )

        return Ubi(
            ubi_obj.min_io_size,
            ubi_obj.leb_size,
            ubi_obj.peb_size,
            ubi_obj.block_count,
            image.image_seq,
            ubi_image_vols,
        )


class UbiUnpacker(Unpacker[None]):
    """
    Extract the UBI image
    """

    targets = (Ubi,)
    children = (UbiVolume,)
    external_dependencies = (PY_LZO_TOOL,)

    async def unpack(self, resource: Resource, config=None):
        with tempfile.TemporaryDirectory() as temp_flush_dir:
            # flush to disk
            with open(f"{temp_flush_dir}/input.img", "wb") as temp_file:
                resource_data = await resource.get_data()
                temp_file.write(resource_data)
                temp_file.flush()

            # extract temp_file to temp_flush_dir
            cmd = [
                "ubireader_extract_images",
                "-o",
                f"{temp_flush_dir}/output",
                temp_file.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)

            ubi_view = await resource.view_as(Ubi)

            # Each file extracted by `ubireader_extract_images` is populated as an UbiVolume
            # `ubireader_extract_images` incorrectly appends a `ubifs` suffix despite unpacking ubi images / volumes
            for vol in ubi_view.volumes:
                f_path = (
                    f"{temp_flush_dir}/output/{os.path.basename(temp_file.name)}"
                    f"/img-{ubi_view.image_seq}_vol-{vol.name}.ubifs"
                )
                with open(f_path, "rb") as f:
                    data = f.read()
                    if len(data) > 0:
                        other_tags: Tuple[ResourceTag, ...] = (GenericBinary,)
                    else:
                        other_tags = ()
                    await resource.create_child_from_view(
                        vol, data=data, additional_tags=other_tags
                    )


class UbiPacker(Packer[None]):
    """
    Generate an UBI image from an Ubi resource view.
    """

    targets = (Ubi,)
    external_dependencies = (UBINIZE_TOOL, PY_LZO_TOOL)

    async def pack(self, resource: Resource, config=None) -> None:
        ubi_view = await resource.view_as(Ubi)

        # with tempfile.NamedTemporaryFile(mode="rb") as temp:
        with tempfile.TemporaryDirectory() as temp_flush_dir:
            ubi_volumes = await resource.get_children()
            ubinize_ini_entries = []

            for volume in ubi_volumes:
                volume_view = await volume.view_as(UbiVolume)
                volume_size = await volume.get_data_length()

                # I think the `ubinize` rounds up the number of required PEBs based on the provided size.
                # Maybe this? allocated PEBs = -(volume_size // -peb_size) + 1
                # For empty volumes I reverse this operation
                if volume_size != 0:
                    volume_path = (
                        f"{temp_flush_dir}/input-{ubi_view.image_seq}_vol-{volume_view.name}.ubivol"
                    )
                    await volume.flush_to_disk(volume_path)
                else:
                    volume_path = None
                    volume_size = (volume_view.peb_count - 1) * ubi_view.peb_size

                # Generate a volume entry for `ubinize`'s config.ini
                ubinize_ini_entry = f"""\
[{volume_view.name}-volume]
mode=ubi
{(f"image={volume_path}" if volume_path else "")}
vol_id={volume_view.id}
vol_size={volume_size}
vol_type={volume_view.type}
vol_name={volume_view.name}
{(f"vol_flags=autoresize" if volume_view.flag_autoresize else "" )}
"""
                ubinize_ini_entries.append(ubinize_ini_entry)

            with open(f"{temp_flush_dir}/config.ini", "w") as config_ini_file:
                config_ini_file.write("\n".join(ubinize_ini_entries))

            cmd = [
                "ubinize",
                "-p",
                str(ubi_view.peb_size),
                "-m",
                str(ubi_view.min_io_size),
                "-o",
                f"{temp_flush_dir}/output.ubi",
                f"{temp_flush_dir}/config.ini",
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)

            with open(f"{temp_flush_dir}/output.ubi", "rb") as output_f:
                packed_blob_data = output_f.read()

            resource.queue_patch(Range(0, await resource.get_data_length()), packed_blob_data)


class UbiIdentifier(Identifier):
    """
    Check the first four bytes of a resource and tag the resource as Ubi if it matches the file magic.
    """

    targets = (File, GenericBinary)

    external_dependencies = (PY_LZO_TOOL,)

    async def identify(self, resource: Resource, config=None) -> None:
        datalength = await resource.get_data_length()
        if datalength >= 4:
            data = await resource.get_data(Range(0, 4))
            if data in [UBI_EC_HDR_MAGIC, UBI_VID_HDR_MAGIC]:
                resource.add_tag(Ubi)



import asyncio
import tempfile
from dataclasses import dataclass
from typing import Optional
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.model.component_model import ComponentConfig, ComponentExternalTool
from ofrak.resource import Resource
from ofrak_type.range import Range

ZSTD = ComponentExternalTool(
    "zstd", "http://facebook.github.io/zstd/", "--help", apt_package="zstd", brew_package="zstd"
)


class ZstdData(GenericBinary):
    """
    A zstd binary blob.
    """

    async def get_child(self) -> GenericBinary:
        return await self.resource.get_only_child_as_view(GenericBinary)


@dataclass
class ZstdPackerConfig(ComponentConfig):
    compression_level: int


class ZstdUnpacker(Unpacker[None]):
    """
    Unpack (decompress) a zstd file.
    """

    id = b"ZstdUnpacker"
    targets = (ZstdData,)
    children = (GenericBinary,)
    external_dependencies = (ZSTD,)

    async def unpack(self, resource: Resource, config: ComponentConfig = None) -> None:
        with tempfile.NamedTemporaryFile(suffix=".zstd") as compressed_file:
            compressed_file.write(await resource.get_data())
            compressed_file.flush()
            output_filename = tempfile.mktemp()

            cmd = [
                "zstd",
                "-d",
                "-k",
                compressed_file.name,
                "-o",
                output_filename,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)
            with open(output_filename, "rb") as f:
                result = f.read()

            await resource.create_child(tags=(GenericBinary,), data=result)


class ZstdPacker(Packer[ZstdPackerConfig]):
    """
    Pack data into a compressed zstd file.
    """

    targets = (ZstdData,)
    external_dependencies = (ZSTD,)

    async def pack(self, resource: Resource, config: Optional[ZstdPackerConfig] = None):
        if config is None:
            config = ZstdPackerConfig(compression_level=19)
        zstd_view = await resource.view_as(ZstdData)
        child_file = await zstd_view.get_child()
        uncompressed_data = await child_file.resource.get_data()

        with tempfile.NamedTemporaryFile() as uncompressed_file:
            uncompressed_file.write(uncompressed_data)
            uncompressed_file.flush()
            output_filename = tempfile.mktemp()

            command = ["zstd", "-T0", f"-{config.compression_level}"]
            if config.compression_level > 19:
                command.append("--ultra")
            command.extend([uncompressed_file.name, "-o", output_filename])
            proc = await asyncio.create_subprocess_exec(
                *command,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=command)
            with open(output_filename, "rb") as f:
                result = f.read()

            compressed_data = result
            original_size = await zstd_view.resource.get_data_length()
            resource.queue_patch(Range(0, original_size), compressed_data)


MagicMimeIdentifier.register(ZstdData, "application/x-zstd")
MagicDescriptionIdentifier.register(
    ZstdData, lambda s: s.lower().startswith("zstandard compressed data")
)



from abc import ABC
from dataclasses import dataclass
from typing import Optional, Tuple


from ofrak.component.analyzer import Analyzer
from ofrak.component.modifier import Modifier
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.memory_region import MemoryRegion
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import index, ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.assembler.assembler_service_i import AssemblerServiceInterface
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface
from ofrak_type.architecture import InstructionSetMode
from ofrak_type.range import Range


@dataclass
class Instruction(MemoryRegion):
    """
    A single ISA instruction.

    :ivar virtual_address: the virtual address of the start of the instruction
    :ivar size: the size of the instruction
    :ivar disassembly: the instruction's disassembly
    :ivar mnemonic: the instruction's mnemonic
    :ivar operands: the instruction's operands
    :ivar mode: the instruction set mode of the instruction
    """

    disassembly: str
    mnemonic: str
    operands: str
    mode: InstructionSetMode

    @index
    def Mnemonic(self) -> str:
        return self.mnemonic

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            instruction_attributes = all_attributes[AttributesType[Instruction]]
        except KeyError:
            return super().caption(all_attributes)
        return f"{instruction_attributes.mnemonic} {instruction_attributes.operands}"

    async def get_assembly(self) -> str:
        """
        Get the instruction as an assembly string.

        :return: the instruction as an assembly string
        """
        return f"{self.mnemonic} {self.operands}"

    async def modify_assembly(
        self,
        mnemonic: Optional[str] = None,
        operands: Optional[str] = None,
        mode: Optional[InstructionSetMode] = None,
    ) -> bytes:
        """
        Modify the instruction, changing it to an instruction of equal size.

        :param mnemonic: the modified instruction mnemonic
        :param operands: the modified instruction operands
        :param mode: the modified instruction's instruction set mode

        :return: the instruction's machine code after modification
        """
        modification_config = InstructionModifierConfig(
            mnemonic or self.mnemonic,
            operands or self.operands,
            mode or self.mode,
        )
        await self.resource.run(InstructionModifier, modification_config)
        data_after_modification = await self.resource.get_data()

        return data_after_modification


@dataclass
class RegisterUsage(ResourceAttributes):
    """
    Information about the register usage in a Resource containing some assembly, such as an
    Instruction.

    :ivar registers_read: registers read from when the assembly executes
    :ivar registers_written: registers written to when the assembly executes
    """

    registers_read: Tuple[str, ...]
    registers_written: Tuple[str, ...]


class InstructionAnalyzer(Analyzer[None, Instruction], ABC):
    """
    Analyze an [instruction][ofrak.core.instruction.Instruction] and extract its attributes.
    """

    id = b"InstructionAnalyzer"
    targets = (Instruction,)
    outputs = (Instruction,)


class InstructionRegisterUsageAnalyzer(Analyzer[None, RegisterUsage], ABC):
    """
    Analyze an [instruction][ofrak.core.instruction.Instruction] and extract the list of registers
    read and written.
    """

    id = b"InstructionRegisterUsageAttributesAnalyzer"
    targets = (Instruction,)
    outputs = (RegisterUsage,)


@dataclass
class InstructionModifierConfig(ComponentConfig):
    """
    Config for the [InstructionModifier][ofrak.core.instruction.InstructionModifier].

    :ivar mnemonic: the modified instruction's mnemonic
    :ivar operands: the modified instruction's operands
    :ivar mode: the modified instruction's instruction set mode
    """

    mnemonic: str
    operands: str
    mode: InstructionSetMode


class InstructionModifier(Modifier[InstructionModifierConfig]):
    """
    Modify an instruction.
    """

    targets = (Instruction,)

    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        assembler_service: AssemblerServiceInterface,
    ):
        super().__init__(
            resource_factory,
            data_service,
            resource_service,
        )
        self._assembler_service = assembler_service

    async def modify(self, resource: Resource, config: InstructionModifierConfig):
        """
        Modify an instruction.

        :param resource: the instruction resource to modify
        :param config:

        :raises AssertionError: if the modified instruction length does not match the length of
        the original instruction
        """
        resource_memory_region = await resource.view_as(MemoryRegion)

        modified_assembly = f"{config.mnemonic} {config.operands}"

        asm = await self._assembler_service.assemble(
            modified_assembly,
            resource_memory_region.virtual_address,
            await resource.analyze(ProgramAttributes),
            config.mode,
        )
        assert (
            len(asm) == resource_memory_region.size
        ), "The modified instruction length does not match the original instruction length"

        new_attributes = AttributesType[Instruction](
            disassembly=modified_assembly,
            mnemonic=config.mnemonic,
            operands=config.operands,
            mode=config.mode,
        )
        resource.queue_patch(Range.from_size(0, len(asm)), asm)
        resource.add_attributes(new_attributes)



from typing import Any

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class AnySerializer(SerializerInterface):
    """
    Serialize and deserialize objects with type hint `Any` into `PJSONType`.

    Implementation: a simple and conservative design in which the object is only serialized
    if it's already of `PJSONType` (the check is performed by beartype and the method signatures).
    """

    targets = (Any,)  # type: ignore

    def obj_to_pjson(self, obj: PJSONType, _type_hint: Any) -> PJSONType:
        return obj

    def pjson_to_obj(self, pjson_obj: PJSONType, _type_hint: Any) -> PJSONType:
        return pjson_obj



import asyncio
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Type, TypeVar

from ofrak.model.data_model import DataPatch
from ofrak.model.resource_model import ResourceAttributes
from ofrak_type.range import Range

CLIENT_COMPONENT_ID = b"__client_context__"
CLIENT_COMPONENT_VERSION = -1


@dataclass
class ComponentConfig:
    """
    Base class for all components' configs. All subclasses should also be dataclasses.
    """


@dataclass(frozen=True)
class ComponentExternalTool:
    """
    An external tool or utility (like `zip` or `squashfs`) a component depends on. Includes some
    basic information on installation, either via package manager or bespoke process.

    Part of this class's responsibility is to check if the tool is installed. Most tools are
    simple command-line utilities whose installation can be check by running:
        `<tool> <install_check_arg>`
    For dependencies which do NOT follow this pattern, subclass ComponentExternalTool and redefine
    the `is_tool_installed` method to perform the check.

    :ivar tool: Name of the command-line tool that will be run
    :ivar tool_homepage: Like to homepage of the tool, with install instructions etc.
    :ivar install_check_arg: Argument to pass to the tool to check if it can be found and run on
    the host, typically something like "--help"
    :ivar apt_package: An `apt` package that installs this tool, if such a package exists
    :ivar brew_package: An `brew` package that installs this tool, if such a package exists

    """

    tool: str
    tool_homepage: str
    install_check_arg: str
    apt_package: Optional[str] = None
    brew_package: Optional[str] = None

    async def is_tool_installed(self) -> bool:
        """
        Check if a tool is installed by running it with the `install_check_arg`.
        This method runs `<tool> <install_check_arg>`.

        :return: True if the `tool` command returned zero, False if `tool` could not be found or
        returned non-zero exit code.
        """
        try:
            cmd = [
                self.tool,
                self.install_check_arg,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )

            returncode = await proc.wait()
        except FileNotFoundError:
            return False

        return 0 == returncode


CC = TypeVar("CC", bound=Optional[ComponentConfig])


@dataclass
class ComponentRunResult:
    """
    Dataclass created after one or more components complete, holding high-level information about
    what resources were affected by a component or components
    """

    components_run: Set[bytes] = field(default_factory=set)
    resources_modified: Set[bytes] = field(default_factory=set)
    resources_deleted: Set[bytes] = field(default_factory=set)
    resources_created: Set[bytes] = field(default_factory=set)

    def update(self, other_results: "ComponentRunResult"):
        self.components_run.update(other_results.components_run)
        self.resources_modified.update(other_results.resources_modified)
        self.resources_created.update(other_results.resources_created)
        self.resources_deleted.update(other_results.resources_deleted)


@dataclass
class ComponentResourceAccessTracker:
    data_accessed: Set[Range] = field(default_factory=set)
    attributes_accessed: Set[Type[ResourceAttributes]] = field(default_factory=set)


@dataclass
class ComponentResourceModificationTracker:
    data_patches: List[DataPatch] = field(default_factory=list)


@dataclass
class ComponentContext:
    component_id: bytes
    component_version: int
    access_trackers: Dict[bytes, ComponentResourceAccessTracker] = field(
        default_factory=lambda: defaultdict(ComponentResourceAccessTracker)
    )
    modification_trackers: Dict[bytes, ComponentResourceModificationTracker] = field(
        default_factory=lambda: defaultdict(ComponentResourceModificationTracker)
    )
    resources_created: Set[bytes] = field(default_factory=set)
    resources_deleted: Set[bytes] = field(default_factory=set)

    def mark_resource_modified(self, r_id: bytes):
        # Creates a new tracker if none exists, and leaves tracker untouched if it already exists
        _ = self.modification_trackers[r_id]

    def get_modified_resource_ids(self, include_deleted=False) -> Set[bytes]:
        modified_resource_ids = set(self.modification_trackers.keys())
        if not include_deleted:
            modified_resource_ids = modified_resource_ids.difference(self.resources_deleted)
        return modified_resource_ids


class ClientComponentContext(ComponentContext):
    def __init__(self):
        super().__init__(CLIENT_COMPONENT_ID, CLIENT_COMPONENT_VERSION)



import asyncio
import logging
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType

from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier

from ofrak.core.binary import GenericBinary
from ofrak.model.component_model import ComponentExternalTool
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

MKSQUASHFS = ComponentExternalTool(
    "mksquashfs", "https://github.com/plougher/squashfs-tools.git", "-help"
)


class _UnsquashfsV45Tool(ComponentExternalTool):
    def __init__(self):
        super().__init__("unsquashfs", "https://github.com/plougher/squashfs-tools.git", "")

    async def is_tool_installed(self) -> bool:
        try:
            cmd = ["unsquashfs", "-help"]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            stdout, stderr = await proc.communicate()
        except FileNotFoundError:
            return False

        if 0 != proc.returncode:
            return False

        if b"-no-exit" not in stdout:
            # Version 4.5+ has the required -no-exit option
            return False

        return True


UNSQUASHFS = _UnsquashfsV45Tool()


@dataclass
class SquashfsFilesystem(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a squashfs format.
    """


class SquashfsUnpacker(Unpacker[None]):
    """Unpack a SquashFS filesystem."""

    targets = (SquashfsFilesystem,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (UNSQUASHFS,)

    async def unpack(self, resource: Resource, config=None):
        with tempfile.NamedTemporaryFile() as temp_file:
            resource_data = await resource.get_data()
            temp_file.write(resource_data)
            temp_file.flush()

            with tempfile.TemporaryDirectory() as temp_flush_dir:
                cmd = [
                    "unsquashfs",
                    "-no-exit-code",
                    "-force",
                    "-dest",
                    temp_flush_dir,
                    temp_file.name,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                )
                returncode = await proc.wait()
                if proc.returncode:
                    raise CalledProcessError(returncode=returncode, cmd=cmd)

                squashfs_view = await resource.view_as(SquashfsFilesystem)
                await squashfs_view.initialize_from_disk(temp_flush_dir)


class SquashfsPacker(Packer[None]):
    """
    Pack files into a compressed squashfs filesystem.
    """

    targets = (SquashfsFilesystem,)
    external_dependencies = (MKSQUASHFS,)

    async def pack(self, resource: Resource, config=None):
        squashfs_view: SquashfsFilesystem = await resource.view_as(SquashfsFilesystem)
        temp_flush_dir = await squashfs_view.flush_to_disk()
        with tempfile.NamedTemporaryFile(suffix=".sqsh", mode="rb") as temp:
            cmd = [
                "mksquashfs",
                temp_flush_dir,
                temp.name,
                "-noappend",
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)
            new_data = temp.read()
            # Passing in the original range effectively replaces the original data with the new data
            resource.queue_patch(Range(0, await resource.get_data_length()), new_data)


MagicMimeIdentifier.register(SquashfsFilesystem, "application/filesystem+sqsh")
MagicDescriptionIdentifier.register(
    SquashfsFilesystem, lambda s: s.startswith("Squashfs filesystem")
)



from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.modifier import Modifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.model.resource_model import ResourceAttributes, ResourceModel, Data
from ofrak.model.tag_model import ResourceTag
from ofrak.ofrak_context import OFRAK, OFRAKContext
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeValueFilter,
    ResourceAttributeRangeFilter,
    ResourceAttributeFilter,
    ResourceAttributeValuesFilter,
    ResourceFilterCondition,
    ResourceSort,
)



from abc import ABC, abstractmethod
from typing import Optional

from ofrak.component.abstract import AbstractComponent
from ofrak.model.component_model import CC
from ofrak.resource import Resource


class ModifierError(RuntimeError):
    pass


class Modifier(AbstractComponent[CC], ABC):
    """
    Modifiers are components that operate on the current state of a resource, directly
    manipulating the underlying binary data.
    """

    @abstractmethod
    async def modify(self, resource: Resource, config: CC) -> None:
        """
        Modify the given resource.

        Users should not call this method directly; rather, they should run
        [Resource.run][ofrak.resource.Resource.run].

        :param resource:
        :param config: Optional config for modification. If an implementation provides a default,
        this default will always be used when config would otherwise be None. Note that a copy of
        the default config will be passed, so the default config values cannot be modified
        persistently by a component run.
        """
        raise NotImplementedError()

    @classmethod
    def get_default_config(cls) -> Optional[CC]:
        return cls._get_default_config_from_method(cls.modify)

    async def _run(self, resource: Resource, config: CC) -> None:
        await self.modify(resource, config)



import asyncio
from typing import Optional, Dict, Type, Tuple

from ofrak.model.tag_model import ResourceTag

from ofrak.component.unpacker import Unpacker
from ofrak.core.code_region import CodeRegion
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter
from ofrak.core.elf.model import (
    ElfHeader,
    ElfProgramHeader,
    ElfProgramHeaderType,
    ElfSegment,
    ElfSectionHeader,
    ElfSection,
    ElfStringSection,
    ElfSectionType,
    ElfSymbolSection,
    ElfSymbol,
    ElfBasicHeader,
    ElfSectionFlag,
    ElfSectionNameStringSection,
    Elf,
    ElfSegmentStructure,
    ElfSectionStructure,
    ElfRelaSection,
    ElfRelaEntry,
    ElfDynSymbolSection,
    ElfDynamicSection,
    ElfDynamicEntry,
    ElfFiniArraySection,
    ElfInitArraySection,
    ElfPointerArraySection,
    ElfSymbolStructure,
    ElfVirtualAddress,
)
from ofrak_type.bit_width import BitWidth
from ofrak_type.memory_permissions import MemoryPermissions
from ofrak_type.range import Range


class ElfUnpacker(Unpacker[None]):
    id = b"ElfUnpacker"
    targets = (Elf,)
    children = (
        ElfBasicHeader,
        ElfHeader,
        ElfProgramHeader,
        ElfSegment,
        ElfSectionHeader,
        ElfSection,
        ElfStringSection,
        ElfSymbolSection,
        CodeRegion,
    )

    async def unpack(self, resource: Resource, config=None):
        """
        Unpack ELF headers and sections / segments into the OFRAK resource tree.

        After unpacking the ElfHeader, the unpacker first unpacks ElfSections defined in the ElfSectionHeader into the
        resource tree.

        If no executable sections are found, then the proceeding ElfProgramHeader unpacking routine is permitted to
        unpack LOAD-able ElfSegments into the resource tree. Executable LOAD-able ElfSegments are unpacked into
        CodeRegions.

        legend:  +    composable child
                 ^    exclusive child
                [ ]   optional / conditional child
                (s)   one or more instances of child

        RESOURCE
         + ElfBasicHeader
         + ElfHeader
           [+] ElfProgramHeader              (creates LOAD-able ElfSegments if ElfSectionHeaders yields no CodeRegions)
           [+] ElfSectionHeader
        [+] ElfSection(s)
           [^] ElfFiniArraySection           (if FINI_ARRAY)
           [^] ElfInitArraySection           (if INIT_ARRAY)
           [^] ElfDynamicSection             (if DYNAMIC)
           [^] ElfRelaSection                (if RELA)
           [^] ElfDynSymbolSection           (if DYNSYM)
           [^] ElfSymbolSection              (if SYMTAB)
           [^] ElfStringSection              (if STRTAB)
           [+] CodeRegion                    (if executable)
           [+] ElfSectionNameStringSection   (if name string section)
        [+] ElfSegment(s)                    (if ElfSection yields no CodeRegions,
           [+] CodeRegion (if executable)     LOAD-able ElfSegments will populate the body of the resource tree instead)
        """

        e_basic_header_r = await resource.create_child(
            tags=(ElfBasicHeader,), data_range=Range(0, 16)
        )
        e_basic_header = await e_basic_header_r.view_as(ElfBasicHeader)

        e_header_range = Range.from_size(
            16, 36 if e_basic_header.get_bitwidth() is BitWidth.BIT_32 else 48
        )
        e_header_r = await resource.create_child(tags=(ElfHeader,), data_range=e_header_range)
        e_header = await e_header_r.view_as(ElfHeader)

        ###########################################################################################
        ### Unpack section headers and associated sections

        sections_by_range_start: Dict[int, Resource] = dict()
        code_region_present = None

        # Create the section header/body resources
        for index in range(e_header.e_shnum):
            e_section_header_offset = e_header.e_shoff + index * e_header.e_shentsize
            e_section_header_range = Range.from_size(e_section_header_offset, e_header.e_shentsize)
            e_section_header_r = await resource.create_child(
                tags=(ElfSectionHeader,),
                data_range=e_section_header_range,
                attributes=(AttributesType[ElfSectionStructure](index),),
            )
            e_section_header = await e_section_header_r.view_as(ElfSectionHeader)

            e_section_offset = e_section_header.sh_offset
            e_section_range = Range.from_size(e_section_offset, e_section_header.sh_size)
            opt_e_section_range: Optional[Range] = e_section_range
            if e_section_range.length() == 0:
                # It's possible that the section does not contain any data
                opt_e_section_range = None
            if e_section_header.get_type() is ElfSectionType.NOBITS:
                # NOBITS sections never have data in the file; their sh_size refers to in-mem size
                opt_e_section_range = None
            e_section_r = await resource.create_child(
                tags=(ElfSection,),
                data_range=opt_e_section_range,
                attributes=(AttributesType[ElfSectionStructure](index),),
            )
            sections_by_range_start[e_section_offset] = e_section_r
            if e_section_header.get_type() is ElfSectionType.FINI_ARRAY:
                e_section_r.add_tag(ElfFiniArraySection)
            elif e_section_header.get_type() is ElfSectionType.INIT_ARRAY:
                e_section_r.add_tag(ElfInitArraySection)
            elif e_section_header.get_type() is ElfSectionType.DYNAMIC:
                e_section_r.add_tag(ElfDynamicSection)
            elif e_section_header.get_type() is ElfSectionType.RELA:
                e_section_r.add_tag(ElfRelaSection)
            elif e_section_header.get_type() is ElfSectionType.DYNSYM:
                e_section_r.add_tag(ElfDynSymbolSection)
            elif e_section_header.get_type() is ElfSectionType.SYMTAB:
                e_section_r.add_tag(ElfSymbolSection)
            elif e_section_header.get_type() is ElfSectionType.STRTAB:
                e_section_r.add_tag(ElfStringSection)

            if e_section_header.has_flag(ElfSectionFlag.EXECINSTR):
                e_section_r.add_tag(CodeRegion)
                code_region_present = True

            if index == e_header.e_shstrndx:
                e_section_r.add_tag(ElfSectionNameStringSection)

        ###########################################################################################
        ### Unpack segment headers, conditionally unpack associated loadable segments

        # Create the program header resources
        for index in range(e_header.e_phnum):
            e_program_header_offset = e_header.e_phoff + (index * e_header.e_phentsize)
            e_program_header_range = Range.from_size(e_program_header_offset, e_header.e_phentsize)

            e_program_header_r = await resource.create_child(
                tags=(ElfProgramHeader,),
                data_range=e_program_header_range,
                attributes=(AttributesType[ElfSegmentStructure](index),),
            )

            e_program_header = await e_program_header_r.view_as(ElfProgramHeader)

            # Don't unpack loadable segments if the section unpacker had already unpacked any CodeRegions
            if code_region_present:
                continue

            if e_program_header.p_type == ElfProgramHeaderType.LOAD.value:
                e_segment_offset = e_program_header.p_offset
                e_segment_range = Range.from_size(e_segment_offset, e_program_header.p_filesz)
                opt_e_segment_range: Optional[Range] = e_segment_range

                # Loaded segments can have no data to initialize with (heap, bss, etc.)
                if e_segment_range.length() == 0:
                    opt_e_segment_range = None

                # We need to inform OFRAK data service the order of child nodes to be populated into the resource tree,
                # since there may be overlapping regions due to flattening non-flat memory structures derived from the
                # binary-under-analysis. We should deprecate `data_after` and `data_before` in favor of a structure
                # that can hold multiple memory views (virtual, physical, file, etc.) of an analyzed binary.
                e_segment_r = await resource.create_child(
                    tags=(ElfSegment,),
                    data_range=opt_e_segment_range,
                    attributes=(AttributesType[ElfSegmentStructure](index),),
                )

                # Tag the segment as a CodeRegion if the loaded segment is executable
                memory_permissions = e_program_header.get_memory_permissions()
                if memory_permissions & MemoryPermissions.R is MemoryPermissions.R:
                    e_segment_r.add_tag(CodeRegion)
                await e_segment_r.save()


class ElfDynamicSectionUnpacker(Unpacker[None]):
    id = b"ElfDynamicSectionUnpacker"
    targets = (ElfDynamicSection,)
    children = (ElfDynamicEntry,)

    async def unpack(self, resource: Resource, config=None):
        e_section = await resource.view_as(ElfDynamicSection)
        elf_r = await e_section.get_elf()
        e_basic_header = await elf_r.get_basic_header()
        dyn_entry_size = 16 if e_basic_header.get_bitwidth() is BitWidth.BIT_64 else 8
        await make_children_helper(resource, ElfDynamicEntry, dyn_entry_size, None)


class ElfRelaUnpacker(Unpacker[None]):
    id = b"ElfRelaUnpacker"
    targets = (ElfRelaSection,)
    children = (ElfRelaEntry,)

    async def unpack(self, resource: Resource, config=None):
        e_section = await resource.view_as(ElfRelaSection)
        elf_r = await e_section.get_elf()
        e_basic_header = await elf_r.get_basic_header()
        rela_size = 24 if e_basic_header.get_bitwidth() is BitWidth.BIT_64 else 12
        await make_children_helper(resource, ElfRelaEntry, rela_size, None)


class ElfSymbolUnpacker(Unpacker[None]):
    id = b"ElfSymbolUnpacker"
    targets = (ElfSymbolSection,)
    children = (ElfSymbol,)

    async def unpack(self, resource: Resource, config=None):
        e_section = await resource.view_as(ElfSymbolSection)
        elf_r = await e_section.get_elf()
        e_basic_header = await elf_r.get_basic_header()
        symbol_size = 16 if e_basic_header.get_bitwidth() is BitWidth.BIT_32 else 24
        await make_children_helper(
            resource, ElfSymbol, symbol_size, AttributesType[ElfSymbolStructure]
        )


class ElfPointerArraySectionUnpacker(Unpacker[None]):
    id = b"ElfPointerArraySectionUnpacker"
    targets = (ElfPointerArraySection,)
    children = (ElfVirtualAddress,)

    async def unpack(self, resource: Resource, config=None):
        elf_r = await resource.get_only_ancestor_as_view(Elf, ResourceFilter.with_tags(Elf))
        e_basic_header = await elf_r.get_basic_header()
        addr_size = 4 if e_basic_header.get_bitwidth() is BitWidth.BIT_32 else 8
        await make_children_helper(resource, ElfVirtualAddress, addr_size, None)


async def make_children_helper(
    resource: Resource,
    entry_type: ResourceTag,
    entry_size: int,
    structure_index_type: Optional[Type[ResourceAttributes]],
) -> None:
    elf_section_size = await resource.get_data_length()
    create_child_tasks = []
    for i, offset in enumerate(range(0, elf_section_size, entry_size)):
        if structure_index_type is not None:
            attrs: Tuple[ResourceAttributes, ...] = (structure_index_type(i),)  # type: ignore
        else:
            attrs = ()
        create_child_tasks.append(
            resource.create_child(
                tags=(entry_type,),
                data_range=Range.from_size(offset, entry_size),
                attributes=attrs,
            )
        )
    await asyncio.gather(*create_child_tasks)



from dataclasses import dataclass

from ofrak.resource import Resource

from ofrak.resource_view import ResourceView
from ofrak_type.range import Range
from ofrak.component.modifier import Modifier, ModifierError
from ofrak.model.component_model import ComponentConfig


class GenericBinary(ResourceView):
    """
    A generic binary blob.
    """


class GenericText(GenericBinary):
    """
    A binary that consists of lines of text.
    """


@dataclass
class BinaryExtendConfig(ComponentConfig):
    """
    Config for the [BinaryExtendModifier][ofrak.core.binary.BinaryExtendModifier].

    :ivar content: the extended bytes.
    """

    content: bytes


class BinaryExtendModifier(Modifier[BinaryExtendConfig]):
    """
    Extend a binary with new data.
    """

    targets = ()

    async def modify(self, resource: Resource, config: BinaryExtendConfig):
        if len(config.content) == 0:
            raise ValueError("Content of the extended space not provided")
        data = await resource.get_data()
        data += config.content
        resource.queue_patch(Range(0, await resource.get_data_length()), data)


@dataclass
class BinaryPatchConfig(ComponentConfig):
    """
    Config for the [BinaryPatchModifier][ofrak.core.binary.BinaryPatchModifier].

    :ivar offset: physical offset from beginning of resource where the patch should be applied
    :ivar patch_bytes: the raw bytes to patch
    """

    offset: int
    patch_bytes: bytes

    def get_range(self):
        return Range.from_size(self.offset, len(self.patch_bytes))


class BinaryPatchModifier(Modifier[BinaryPatchConfig]):
    """
    Patch a binary at the target offset with the given patch bytes.
    """

    targets = ()

    async def modify(self, resource: Resource, config: BinaryPatchConfig) -> None:
        """
        Patch the resource at the target offset with the given patch bytes.

        :param resource: the resource to patch
        :param config: contains the offset at which to patch and the patch bytes to apply

        :raises ModifierError: if the binary patch overflows the original size of the resource
        """
        resource_size = await resource.get_data_length()
        if len(config.patch_bytes) > resource_size - config.offset:
            raise ModifierError(
                f"The binary patch, {config}, overflows the original size of the resource "
                f"{resource.get_id().hex()}."
            )
        resource.queue_patch(config.get_range(), config.patch_bytes)
        return



import logging
from abc import ABC, abstractmethod
from typing import Type, Tuple, Generic, TypeVar, Union, cast, Iterable, Optional

from ofrak.resource import Resource

from ofrak.component.abstract import AbstractComponent
from ofrak.model.component_model import CC
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import ViewableResourceTag
from ofrak.resource_view import ResourceView

LOGGER = logging.getLogger(__name__)

AnalyzerReturnType = TypeVar(
    "AnalyzerReturnType",
    bound=Union[
        Tuple[ResourceAttributes, ...],
        ResourceAttributes,
        ResourceView,
    ],
)


class AnalyzerError(RuntimeError):
    """Base exception raised by analyzers."""


class Analyzer(AbstractComponent, Generic[CC, AnalyzerReturnType], ABC):
    """
    Analyzers are discrete components that gather and analyze specific information from the target
    [Resource][ofrak.resource.Resource]. Analyzers return custom processed
    data results.
    """

    @property
    @abstractmethod
    def outputs(
        self,
    ) -> Union[Tuple[Type[ResourceAttributes]], Tuple[ViewableResourceTag]]:
        raise NotImplementedError()

    @abstractmethod
    async def analyze(self, resource: Resource, config: CC) -> AnalyzerReturnType:
        """
        Analyze a resource for to extract specific
        [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes].

        Users should not call this method directly; rather, they should run
        [Resource.run][ofrak.resource.Resource.run] or
        [Resource.analyze][ofrak.resource.Resource.analyze].

        :param resource: The resource that is being analyzed
        :param config: Optional config for analyzing. If an implementation provides a default,
        this default will always be used when config would otherwise be None. Note that a copy of
        the default config will be passed, so the default config values cannot be modified
        persistently by a component run.
        :return: The analysis results
        """
        raise NotImplementedError()

    def get_attributes_from_results(
        self, results: AnalyzerReturnType
    ) -> Iterable[ResourceAttributes]:
        if isinstance(results, tuple):
            attributes = results
        elif isinstance(results, ResourceAttributes):
            attributes = (results,)
        elif isinstance(results, ResourceView):
            attributes = tuple(results.get_attributes_instances().values())
        else:
            raise TypeError(
                f"Analyzer {type(self).__name__} returned an instance of type"
                f" {type(results)}; expected ResourceAttributes, Tuple[ResourceAttributes], "
                f"or ResourceView."
            )
        return attributes

    def get_outputs_as_attribute_types(self) -> Tuple[Type[ResourceAttributes], ...]:
        # All elements of the self.outputs tuple must either be subclasses of ResourceAttributes,
        # or must have ViewableResourceTag as a metaclass. This should always be true based on
        # the type annotation that is a union of two possible types of homogenous tuples
        all_outputs_are_viewable_tags = all(
            isinstance(output, ViewableResourceTag) for output in self.outputs
        )
        all_outputs_are_resource_attributes = all(
            issubclass(output, ResourceAttributes) for output in self.outputs
        )

        if all_outputs_are_viewable_tags and not all_outputs_are_resource_attributes:
            outputs = []
            for output in self.outputs:
                outputs += list(cast(ViewableResourceTag, output).composed_attributes_types)
            return tuple(outputs)
        elif not all_outputs_are_viewable_tags and all_outputs_are_resource_attributes:
            return cast(Tuple[Type[ResourceAttributes]], self.outputs)
        else:
            raise TypeError(
                f"Analyzer {type(self).__name__} has outputs that are either a tuple of "
                f"non-homogenous types, and/or have an element of a type other than "
                f"ViewableResourceTag and Type[ResourceAttributes]."
            )

    @classmethod
    def get_default_config(cls) -> Optional[CC]:
        return cls._get_default_config_from_method(cls.analyze)

    async def _run(self, resource: Resource, config: CC):
        if resource.has_component_run(self.get_id(), self.get_version()):
            return self._log_component_has_run_warning(resource)
        if resource.has_component_run(self.get_id()):
            self._log_component_has_run_warning(resource)
            raise NotImplementedError(
                "If the component has already run (but on a different "
                "version), we should remove the dependencies"
            )
        analysis_results = await self.analyze(resource, config)
        attributes = self.get_attributes_from_results(analysis_results)
        for attrs in attributes:
            resource.add_attributes(attrs)

        resource.add_component(self.get_id(), self.get_version())
        attribute_types = self.get_outputs_as_attribute_types()
        for attribute_type in attribute_types:
            resource.add_component_for_attributes(self.get_id(), self.get_version(), attribute_type)



from dataclasses import dataclass
from typing import List

from ofrak_type.range import Range


@dataclass
class DataPatch:
    """
    Representation of a binary patch to part of a resource's data.

    :ivar range: The slice of the binary blob to replace with new data (zero-length is allowed)
    :ivar data_id: ID of the binary blob to apply this path to
    :ivar data: The bytes to replace old data with (zero-length is allowed)
    """

    range: Range
    data_id: bytes
    data: bytes

    def __repr__(self):
        return f"DataPatch({self.data_id.hex()}, {self.range}, {len(self.data)})"


@dataclass
class DataPatchesResult:
    """
    Summary of changes to a binary blob resulting from the application of a ``DataPatch``.

    :ivar data_id: ID of the patched blob
    :ivar patches: Range in the original blob which have been altered by applying patches
    """

    data_id: bytes
    patches: List[Range]


@dataclass
class DataModel:
    """
    Representation of a chunk of binary data stored in a `DataServiceInterface` implementation.

    :ivar id: Unique ID for this data model.
    :ivar range: The slice of some underlying binary blob which belongs to this data model. For
    root models, this range starts from 0 and ends at the total length of the blob.
    :ivar root_id: If this model represents a part of a root model's data, this is the ID of that
    root data model.
    """

    id: bytes
    range: Range
    root_id: bytes

    def is_mapped(self):
        return self.root_id != self.id



from ofrak.component.analyzer import Analyzer
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.patch_maker.linkable_symbol import LinkableSymbol, LinkableSymbolType
from ofrak.resource import Resource
from ofrak.model.component_model import ComponentConfig


class ComplexBlockSymbolAnalyzer(Analyzer[None, LinkableSymbol]):
    targets = (ComplexBlock,)
    outputs = (LinkableSymbol,)

    async def analyze(self, resource: Resource, config: ComponentConfig = None) -> LinkableSymbol:
        cb = await resource.view_as(ComplexBlock)
        cb_mode = await cb.get_mode()

        return LinkableSymbol(
            cb.VirtualAddress,
            cb.Symbol,
            LinkableSymbolType.FUNC,
            cb_mode,
        )



from ofrak.resource import ResourceFactory

from ofrak.model.component_model import ComponentExternalTool
from ofrak.service.component_locator_i import ComponentLocatorInterface
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface
from pytest_ofrak.mock_component_types import MockUnpacker


class _MockComponentA(MockUnpacker):
    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        component_locator: ComponentLocatorInterface,
    ):
        super(MockUnpacker, self).__init__(
            resource_factory, data_service, resource_service, component_locator
        )

    external_dependencies = (ComponentExternalTool("tool_a", "tool_a.com", "--help", "tool_a_apt"),)


class _MockComponentB(MockUnpacker):
    external_dependencies = (
        ComponentExternalTool("tool_b", "tool_b.com", "--help", None, "tool_a_brew"),
    )


class _MockComponentC(MockUnpacker):
    pass



from typing import Any, Tuple, List, Union

from ofrak.core.patch_maker.modifiers import SourceBundle
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class SourceBundleSerializer(SerializerInterface):
    """
    Serialize and deserialize `Dict[X, Y]` into `PJSONType`.

    Implementation: a list of tuples (key, value), because the serialized keys may not be hashable
    (e.g. if they're themselves serialized as a dictionary), despite the real keys being hashable.
    """

    targets = (SourceBundle,)

    # Tuples are kept as tuples in PJSON, but serialized into lists in JSON.
    SerializedDictType = Union[List[Tuple[PJSONType, PJSONType]], List[List[PJSONType]]]

    def obj_to_pjson(self, obj: SourceBundle, type_hint: Any) -> SerializedDictType:
        return [
            (self._service.to_pjson(key, str), self._serialize_value(value))
            for key, value in obj.items()
        ]

    def pjson_to_obj(self, pjson_obj: SerializedDictType, type_hint: Any) -> SourceBundle:
        return SourceBundle(
            [
                (self._service.from_pjson(key, str), self._deserialize_value(value))
                for key, value in pjson_obj
            ]
        )

    def _serialize_value(self, value: Union[bytes, SourceBundle]) -> PJSONType:
        if isinstance(value, SourceBundle):
            return self.obj_to_pjson(value, SourceBundle)
        else:
            # TODO: Only serializing as strings for the GUI SourceBundle inputs, which are
            #  strings, instead of bytes as they should be
            string_value: str = value.decode("utf-8")
            return self._service.to_pjson(string_value, str)

    def _deserialize_value(self, value: Any) -> Union[bytes, SourceBundle]:
        if isinstance(value, str):
            # TODO: Only deserializing from strings for the GUI SourceBundle inputs, which are
            #  strings, instead of bytes as they should be
            deserialized_string: str = self._service.from_pjson(value, str)
            return deserialized_string.encode("utf-8")
        else:
            return self._service.from_pjson(value, SourceBundle)



import logging
from types import ModuleType
from typing import List, Type, Dict, TypeVar, Optional, Set

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.interface import ComponentInterface
from ofrak.component.modifier import Modifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.service.component_locator_i import ComponentLocatorInterface, ComponentFilter
from ofrak_type.error import NotFoundError

from functools import lru_cache

LOGGER = logging.getLogger(__name__)
CI = TypeVar("CI", bound="ComponentInterface")


class InvalidComponentError(RuntimeError):
    pass


class TooFewConstraintsError(RuntimeError):
    pass


class ComponentLocator(ComponentLocatorInterface):
    """
    Locates component singletons from their types or `id`.
    """

    COMPONENT_CATEGORIES = (
        Unpacker,
        Packer,
        Analyzer,
        Modifier,
        Identifier,
    )

    def __init__(self):
        self._components_by_category: Dict[Type[ComponentInterface], List[ComponentInterface]] = {
            category: list() for category in ComponentLocator.COMPONENT_CATEGORIES
        }
        self._components_by_id: Dict[bytes, ComponentInterface] = {}
        self._all_components = set(self._components_by_id.values())

    def _resolve_id_conflict(
        self,
        component_a: ComponentInterface,
        component_b: ComponentInterface,
        module_priority: List[ModuleType],
    ) -> ComponentInterface:
        contested_id = component_a.get_id()
        if component_a == component_b:
            return component_a
        component_a_name = type(component_a).__name__
        component_b_name = type(component_b).__name__

        for module in reversed(module_priority):
            component_a_in_module = all(
                m1 == m2
                for m1, m2 in zip(component_a.__module__.split("."), module.__name__.split("."))
            )
            component_b_in_module = all(
                m1 == m2
                for m1, m2 in zip(component_b.__module__.split("."), module.__name__.split("."))
            )

            if component_a_in_module:
                LOGGER.debug(
                    f"Component ID conflict for {contested_id.decode()} between "
                    f"{component_a_name} and {component_b_name} "
                    f"resolved: {component_a_name}'s module is higher "
                    f"priority, {component_b_name} will not be used."
                )
                return component_a
            elif component_b_in_module:
                LOGGER.debug(
                    f"Component ID conflict for {contested_id.decode()} between "
                    f"{component_a_name} and {component_b_name} "
                    f"resolved: {component_b_name}'s module is higher "
                    f"priority, {component_a_name} will not be used."
                )
                return component_b

        raise InvalidComponentError(
            f"A component with the ID {contested_id.decode()} is already "
            f"registered and given module priorities were insufficient to resolve conflict"
        )

    def add_components(
        self,
        components: List[ComponentInterface],
        module_priority: Optional[List[ModuleType]] = None,
    ):
        for component in components:
            if component.get_id() not in self._components_by_id:
                self._components_by_id[component.get_id()] = component
            else:
                if module_priority is None:
                    raise InvalidComponentError(
                        f"A component with the ID {component.get_id().decode()} is already "
                        f"registered"
                    )
                else:
                    contested_id = component.get_id()
                    existing_component = self._components_by_id[contested_id]

                    self._components_by_id[contested_id] = self._resolve_id_conflict(
                        component, existing_component, module_priority
                    )

        for component in self._components_by_id.values():
            component_category = None
            for t_component_type in self._components_by_category.keys():
                if isinstance(component, t_component_type):
                    component_category = t_component_type
                    break
            if component_category is None:
                category_names = [category.__name__ for category in self.COMPONENT_CATEGORIES]
                raise InvalidComponentError(
                    f"The component {component.get_id().decode()} is not one of {', '.join(category_names)}"
                )
            if component.targets is None:
                raise InvalidComponentError(
                    f"The component {component.get_id().decode()} does not define targets"
                )
            if component.get_version() is None:
                raise InvalidComponentError(
                    f"The component {component.get_id().decode()} does not define a version"
                )
            if isinstance(component, Analyzer):
                if component.outputs is None:
                    raise InvalidComponentError(
                        f"The analyzer {component.get_id().decode()} does not define attributes"
                    )

            self._components_by_category[component_category].append(component)
            LOGGER.debug(
                f"Registered component {component.get_id().decode()} as {component_category.__name__}"
            )

        self._all_components = set(self._components_by_id.values())

    def get_by_id(self, component_id: bytes) -> ComponentInterface:
        component = self._components_by_id.get(component_id)
        if component is None:
            raise NotFoundError(
                f"The component with ID {component_id.decode()} has not been registered"
            )
        return component

    def get_by_type(self, component_type: Type[CI]) -> CI:
        return self.get_by_id(component_type.get_id())  # type: ignore

    @lru_cache(maxsize=None)
    def get_components_matching_filter(
        self, component_filter: ComponentFilter
    ) -> Set[ComponentInterface]:
        # naive implementation to start
        # will be improved by caching

        return component_filter.filter(self._all_components)



import logging
from dataclasses import dataclass
from typing import Callable, Dict, Iterable, Union

from ofrak.component.abstract import ComponentMissingDependencyError

try:
    import magic

    MAGIC_INSTALLED = True
except ImportError:
    MAGIC_INSTALLED = False

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.core.binary import GenericBinary, GenericText
from ofrak.core.filesystem import File
from ofrak.model.component_model import ComponentExternalTool
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.tag_model import ResourceTag
from ofrak.resource import Resource
from ofrak_type.error import AlreadyExistError

LOGGER = logging.getLogger(__name__)


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Magic(ResourceAttributes):
    mime: str
    descriptor: str


class _LibmagicDependency(ComponentExternalTool):
    def __init__(self):
        super().__init__(
            "libmagic",
            "https://www.darwinsys.com/file/",
            install_check_arg="",
            apt_package="libmagic1",
            brew_package="libmagic",
        )

        try:
            import magic as _magic

            _LibmagicDependency._magic = _magic
        except ImportError:
            _LibmagicDependency._magic = None

    async def is_tool_installed(self) -> bool:
        return MAGIC_INSTALLED


LIBMAGIC_DEP = _LibmagicDependency()


class MagicAnalyzer(Analyzer[None, Magic]):
    """
    Analyze a binary blob to extract its mimetype and magic description.
    """

    targets = (File, GenericBinary)
    outputs = (Magic,)
    external_dependencies = (LIBMAGIC_DEP,)

    async def analyze(self, resource: Resource, config=None) -> Magic:
        data = await resource.get_data()
        if not MAGIC_INSTALLED:
            raise ComponentMissingDependencyError(self, LIBMAGIC_DEP)
        else:
            magic_mime = magic.from_buffer(data, mime=True)
            magic_description = magic.from_buffer(data)
            return Magic(magic_mime, magic_description)


class MagicMimeIdentifier(Identifier[None]):
    """
    Identify and add the appropriate tag for a given resource based on its mimetype.
    """

    id = b"MagicMimeIdentifier"
    targets = (File, GenericBinary)
    external_dependencies = (LIBMAGIC_DEP,)  # Indirect thru MagicAnalyzer, but worth tagging

    _tags_by_mime: Dict[str, ResourceTag] = dict()

    async def identify(self, resource: Resource, config=None):
        _magic = await resource.analyze(Magic)
        magic_mime = _magic.mime
        tag = MagicMimeIdentifier._tags_by_mime.get(magic_mime)
        if tag is not None:
            resource.add_tag(tag)

    @classmethod
    def register(cls, resource: ResourceTag, mime_types: Union[Iterable[str], str]):
        if isinstance(mime_types, str):
            mime_types = [mime_types]
        for mime_type in mime_types:
            if mime_type in cls._tags_by_mime:
                raise AlreadyExistError(f"Registering already-registered mime type: {mime_type}")
            cls._tags_by_mime[mime_type] = resource


class MagicDescriptionIdentifier(Identifier[None]):
    """
    Identify and add the appropriate tag for a given resource based on its mime description.
    """

    id = b"MagicDescriptionIdentifier"
    targets = (File, GenericBinary)
    external_dependencies = (LIBMAGIC_DEP,)  # Indirect thru MagicAnalyzer, but worth tagging

    _matchers: Dict[Callable, ResourceTag] = dict()

    async def identify(self, resource: Resource, config):
        _magic = await resource.analyze(Magic)
        magic_description = _magic.descriptor
        for matcher, resource_type in self._matchers.items():
            if matcher(magic_description):
                resource.add_tag(resource_type)

    @classmethod
    def register(cls, resource: ResourceTag, matcher: Callable):
        if matcher in cls._matchers:
            raise AlreadyExistError("Registering already-registered matcher")
        cls._matchers[matcher] = resource


MagicMimeIdentifier.register(GenericText, "text/plain")
MagicDescriptionIdentifier.register(
    GenericText, lambda desc: any([("ASCII text" in s) for s in desc.split(", ")])
)

MagicMimeIdentifier.register(GenericBinary, "application/octet-stream")
MagicDescriptionIdentifier.register(GenericBinary, lambda s: s == "data")



import asyncio
import logging
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from typing import (
    Awaitable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    Any,
)

from ofrak.component.unpacker import Unpacker

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.interface import ComponentInterface
from ofrak.component.packer import Packer
from ofrak.model.component_model import CC, ComponentRunResult
from ofrak.model.job_model import (
    JobModel,
    JobRunContext,
    JobRunContextFactory,
)
from ofrak.model.job_request_model import (
    JobAnalyzerRequest,
    JobComponentRequest,
    JobMultiComponentRequest,
)
from ofrak.model.resource_model import EphemeralResourceContextFactory
from ofrak.model.tag_model import ResourceTag
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak.service.component_locator_i import (
    ComponentLocatorInterface,
    ComponentFilter,
)
from ofrak.model.component_filters import (
    ComponentTargetFilter,
    AnalyzerOutputFilter,
    ComponentOrMetaFilter,
    ComponentAndMetaFilter,
    ComponentPrioritySelectingMetaFilter,
    ComponentNotMetaFilter,
    ComponentWhitelistFilter,
    ComponentTypeFilter,
)
from ofrak.service.job_service_i import (
    JobServiceInterface,
    ComponentAutoRunFailure,
)
from ofrak.service.resource_service_i import (
    ResourceServiceInterface,
    ResourceFilter,
    ResourceFilterCondition,
)
from ofrak_type.error import NotFoundError

TargetCache = Dict[ResourceTag, List[ComponentInterface]]
LOGGER = logging.getLogger(__name__)

MAX_CONCURRENT_COMPONENTS = 512

ANALYZERS_FILTER = ComponentTypeFilter(Analyzer)  # type: ignore
IDENTIFIERS_FILTER = ComponentTypeFilter(Identifier)  # type: ignore
UNPACKERS_FILTER = ComponentTypeFilter(Unpacker)  # type: ignore
PACKERS_FILTER = ComponentTypeFilter(Packer)  # type: ignore

M = TypeVar("M")

_RunTaskResultT = Tuple[Union[ComponentRunResult, BaseException], M]


@dataclass
class _ComponentAutoRunRequest:
    target_resource_id: bytes
    component_filter: ComponentFilter


class JobService(JobServiceInterface):
    def __init__(
        self,
        component_locator: ComponentLocatorInterface,
        resource_service: ResourceServiceInterface,
        resource_context_factory: EphemeralResourceContextFactory,
        job_context_factory: JobRunContextFactory,
    ):
        self._job_store: Dict[bytes, JobModel] = dict()
        self._component_locator = component_locator
        self._resource_service = resource_service
        self._resource_context_factory = resource_context_factory
        self._job_context_factory = job_context_factory

        self._active_component_tasks: Dict[Tuple[bytes, bytes], Awaitable[_RunTaskResultT]] = dict()

    async def create_job(self, id: bytes, name: str) -> JobModel:
        model = JobModel(id, name)
        self._job_store[id] = model
        return model

    async def _run_component(
        self,
        metadata: M,
        job_id: bytes,
        resource_id: bytes,
        component: ComponentInterface,
        job_context: JobRunContext,
        config: CC,
    ) -> _RunTaskResultT:
        """
        Run a component, return the result as well as some (optional) metadata (such as the request
        that triggered the component to run). If it raises an error, the error is returned
        as an object rather than being raised.

        Once the component finishes, log it and remove this task from the set of active tasks.
        """
        LOGGER.info(
            f"JOB {job_id.hex()} - Running {component.get_id().decode()} on "
            f"resource {resource_id.hex()}"
        )

        # Create a new resource context for every component
        fresh_resource_context = self._resource_context_factory.create()
        fresh_resource_view_context = ResourceViewContext()
        result: Union[ComponentRunResult, BaseException]
        try:
            result = await component.run(
                job_id,
                resource_id,
                job_context,
                fresh_resource_context,
                fresh_resource_view_context,
                config,
            )
            _log_component_run_result_info(job_id, resource_id, component, result)
        except Exception as e:
            result = e
        component_task_id = (resource_id, component.get_id())
        del self._active_component_tasks[component_task_id]

        return result, metadata

    def _create_run_component_task(
        self,
        metadata: Any,
        job_id: bytes,
        resource_id: bytes,
        component: ComponentInterface,
        job_context: JobRunContext,
        config: CC = None,
    ) -> Awaitable[_RunTaskResultT]:
        component_task_id = (resource_id, component.get_id())
        if component_task_id in self._active_component_tasks:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(
                    f"JOB {job_id.hex()} - Found already running task {component.get_id().decode()}"
                    f" on resource {resource_id.hex()}, awaiting result."
                )
            duplicate_task = self._active_component_tasks[component_task_id]

            return duplicate_task
        else:
            component_task = asyncio.create_task(
                self._run_component(
                    metadata,
                    job_id,
                    resource_id,
                    component,
                    job_context,
                    config,
                )
            )
            self._active_component_tasks[component_task_id] = component_task
            return component_task

    async def run_component(
        self,
        request: JobComponentRequest,
        job_context: Optional[JobRunContext] = None,
    ) -> ComponentRunResult:
        component = self._component_locator.get_by_id(request.component_id)
        if job_context is None:
            job_context = self._job_context_factory.create()
        result, _ = await self._create_run_component_task(
            request,
            request.job_id,
            request.resource_id,
            component,
            job_context,
            request.config,
        )
        if isinstance(result, BaseException):
            raise result
        else:
            return result

    async def run_analyzer_by_attribute(
        self,
        request: JobAnalyzerRequest,
        job_context: Optional[JobRunContext] = None,
    ) -> ComponentRunResult:
        if job_context is None:
            job_context = self._job_context_factory.create()

        target_resource_model = await self._resource_service.get_by_id(request.resource_id)
        # There may be an analyzer that outputs ALL the requested attributes at once
        # If there is (as is usually the case for views), only run that one
        one_analyzer_for_all_attributes: ComponentFilter = ComponentAndMetaFilter(
            AnalyzerOutputFilter(*request.attributes),
            _build_tag_filter(tuple(target_resource_model.get_tags())),
        )
        # Otherwise, look for individual analyzers for each attributes type
        analyzer_for_each_attribute: List[ComponentFilter] = [
            ComponentAndMetaFilter(
                AnalyzerOutputFilter(attr_t),
                _build_tag_filter(tuple(target_resource_model.get_tags())),
            )
            for attr_t in request.attributes
        ]

        component_filter: ComponentFilter = ComponentAndMetaFilter(
            ANALYZERS_FILTER,
            ComponentPrioritySelectingMetaFilter(
                one_analyzer_for_all_attributes,
                ComponentOrMetaFilter(*analyzer_for_each_attribute),
            ),
        )

        components_result = await self._auto_run_components(
            (
                _ComponentAutoRunRequest(
                    request.resource_id,
                    component_filter,
                ),
            ),
            request.job_id,
            job_context,
        )
        if components_result.components_run:
            return components_result
        else:
            raise NotFoundError(
                f"Unable to find any analyzer for attributes "
                f"{', '.join(attr_t.__name__ for attr_t in request.attributes)}"
                f"\nFilter: {component_filter}"
            )

    def _build_target_cache(self, component_filter: ComponentFilter) -> TargetCache:
        components = self._component_locator.get_components_matching_filter(component_filter)
        target_cache = defaultdict(list)
        for component in components:
            if component.targets is None:
                continue
            for target in component.targets:
                target_cache[target].append(component)
        return target_cache

    async def run_components(
        self,
        request: JobMultiComponentRequest,
    ) -> ComponentRunResult:
        resource = await self._resource_service.get_by_id(request.resource_id)
        component_filter = _build_auto_run_filter(request)

        tags_to_target = tuple(resource.tags)
        components_result = ComponentRunResult()
        while len(tags_to_target) > 0:
            job_context = self._job_context_factory.create()
            component_tag_filter = _build_tag_filter(tags_to_target)
            final_filter = ComponentAndMetaFilter(component_filter, component_tag_filter)
            individual_component_results = await self._auto_run_components(
                (
                    _ComponentAutoRunRequest(
                        request.resource_id,
                        final_filter,
                    ),
                ),
                request.job_id,
                job_context,
            )

            components_result.update(individual_component_results)
            resource_tracker = job_context.trackers[request.resource_id]
            tags_added = resource_tracker.tags_added
            tags_to_target = tuple(tags_added)

        return components_result

    async def run_components_recursively(
        self, request: JobMultiComponentRequest
    ) -> ComponentRunResult:
        components_result = ComponentRunResult()
        component_filter = _build_auto_run_filter(request)

        initial_target_resource_models = await self._get_initial_recursive_target_resources(
            request.resource_id, component_filter
        )

        # Create a mock context to match all existing tags
        previous_job_context: JobRunContext = self._job_context_factory.create()
        for existing_resource_model in initial_target_resource_models:
            previous_job_context.trackers[existing_resource_model.id].tags_added.update(
                existing_resource_model.tags
            )
        iterations = 0
        tags_added_count = 1  # initialize just so loop starts

        while tags_added_count > 0:
            job_context = self._job_context_factory.create()
            _run_components_requests = []
            for resource_id, previous_tracker in previous_job_context.trackers.items():
                final_filter = ComponentAndMetaFilter(
                    component_filter,
                    _build_tag_filter(tuple(previous_tracker.tags_added)),
                )
                _run_components_requests.append(
                    _ComponentAutoRunRequest(
                        resource_id,
                        final_filter,
                    )
                )

            iteration_components_result = await self._auto_run_components(
                _run_components_requests,
                request.job_id,
                job_context,
            )
            components_result.update(iteration_components_result)

            tags_added_count = 0
            for resource_id, tracker in job_context.trackers.items():
                if len(tracker.tags_added) > 0:
                    tags_added_count += len(tracker.tags_added)
            previous_job_context = job_context
            LOGGER.info(
                f"Completed iteration {iterations} of run_components_recursively on "
                f"{request.resource_id.hex()}. {len(components_result.resources_modified)} "
                f"resources modified and {tags_added_count} tags added."
            )
            iterations += 1
        return components_result

    async def pack_recursively(
        self,
        job_id: bytes,
        resource_id: bytes,
    ) -> ComponentRunResult:
        packer_filter = PACKERS_FILTER
        target_cache = self._build_target_cache(packer_filter)
        all_components_result = ComponentRunResult()
        if len(target_cache) == 0:
            return all_components_result
        resources = await self._resource_service.get_descendants_by_id(
            resource_id,
            r_filter=ResourceFilter(
                include_self=True,
                tags=tuple(target_cache.keys()),
                tags_condition=ResourceFilterCondition.OR,
            ),
        )
        resources = list(resources)  # we'll need that Iterable more than once
        job_context = self._job_context_factory.create()

        # We want to start with the deepest packers. Packers at the same levels can run
        # concurrently. So we first ask for the relative depth of each returned resource.
        resource_depths = await self._resource_service.get_depths(
            [resource.id for resource in resources]
        )

        resources_by_depth = defaultdict(list)
        for resource, depth in zip(resources, resource_depths):
            resources_by_depth[depth].append(resource)

        for depth in sorted(resources_by_depth.keys(), reverse=True):
            for resource in resources_by_depth[depth]:
                component_filter: ComponentFilter = ComponentAndMetaFilter(
                    PACKERS_FILTER,
                    _build_tag_filter(tuple(resource.get_tags())),
                )

                request = _ComponentAutoRunRequest(
                    resource.id,
                    component_filter,
                )

                component_result = await self._auto_run_components(
                    [request],
                    job_id,
                    job_context,
                )
                n_packers_run = len(component_result.components_run)
                if n_packers_run == 0:
                    all_components_result.update(component_result)
                    break
                if n_packers_run > 1:
                    raise ValueError(f"Multiple packers are targeting resource {resource.id.hex()}")
        return all_components_result

    async def _get_initial_recursive_target_resources(
        self, resource_id: bytes, component_filter: ComponentFilter
    ):
        possible_targets: Set[ResourceTag] = set()
        for component in self._component_locator.get_components_matching_filter(component_filter):
            possible_targets.update(component.targets)
        initial_target_resource_models = await self._resource_service.get_descendants_by_id(
            resource_id,
            r_filter=ResourceFilter(
                include_self=True, tags=possible_targets, tags_condition=ResourceFilterCondition.OR
            ),
        )
        return initial_target_resource_models

    async def _auto_run_components(
        self,
        requests: Iterable[_ComponentAutoRunRequest],
        job_id: bytes,
        job_context: JobRunContext,
    ) -> ComponentRunResult:
        queue: List[Tuple[_ComponentAutoRunRequest, ComponentInterface]] = []
        for request in requests:
            components = self._component_locator.get_components_matching_filter(
                request.component_filter
            )
            if not components:
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug(
                        f"JOB {job_id.hex()} - Found no components to run on "
                        f"{request.target_resource_id.hex()} matching filters "
                        f"{request.component_filter}"
                    )
            else:
                for component in components:
                    queue.append((request, component))

        concurrent_run_tasks: List[Awaitable[_RunTaskResultT]] = list()

        n_tasks_to_add = min(MAX_CONCURRENT_COMPONENTS, len(queue))
        for _ in range(n_tasks_to_add):
            request, component = queue.pop()
            run_component_task = self._create_run_component_task(
                (request, type(component).__name__),
                job_id,
                request.target_resource_id,
                component,
                job_context,
            )
            concurrent_run_tasks.append(run_component_task)

        components_result = ComponentRunResult(set(), set(), set(), set())
        while len(queue) > 0 or len(concurrent_run_tasks) > 0:
            completed, pending = await asyncio.wait(
                concurrent_run_tasks, return_when=asyncio.FIRST_COMPLETED
            )
            LOGGER.debug(
                f"Completed {len(completed)} component run tasks, {len(pending)} pending and "
                f"{len(queue)} still in queue"
            )
            for completed_task in completed:
                component_run_result, component_run_metadata = completed_task.result()
                if isinstance(component_run_result, ComponentRunResult):
                    components_result.update(component_run_result)
                else:
                    component_run_error = cast(BaseException, component_run_result)
                    request_causing_run, component_name = component_run_metadata
                    raise component_run_error from ComponentAutoRunFailure(
                        request_causing_run.target_resource_id,
                        request_causing_run.component_filter,
                        component_name.encode(),
                    )

            concurrent_run_tasks = list(pending)
            n_tasks_to_add = min(len(completed), len(queue))
            for _ in range(n_tasks_to_add):
                request, component = queue.pop()
                run_component_task = self._create_run_component_task(
                    (request, type(component).__name__),
                    job_id,
                    request.target_resource_id,
                    component,
                    job_context,
                )
                concurrent_run_tasks.append(run_component_task)

        return components_result


def _log_component_run_result_info(
    job_id: bytes,
    resource_id: bytes,
    component: ComponentInterface,
    component_result: ComponentRunResult,
    max_ids_to_log: int = 12,
):
    if LOGGER.getEffectiveLevel() > logging.INFO:
        return

    def truncate_id_seq(id_seq) -> Iterable[str]:
        for n, r_id in enumerate(id_seq):
            if n > max_ids_to_log:
                yield f" ... ({len(id_seq) - n} more)"
                return
            yield r_id.hex()

    logging_component_results = []
    if component_result.resources_modified:
        logging_component_results.append(
            f"Modified resources: {','.join(truncate_id_seq(component_result.resources_modified))}"
        )
    if component_result.resources_created:
        logging_component_results.append(
            f"Created resources: {','.join(truncate_id_seq(component_result.resources_created))}"
        )
    if component_result.resources_deleted:
        logging_component_results.append(
            f"Deleted resources: {','.join(truncate_id_seq(component_result.resources_deleted))}"
        )

    if logging_component_results:
        logging_component_results_str = "\n\t" + ("\n\t".join(logging_component_results))
    else:
        logging_component_results_str = ""
    LOGGER.info(
        f"JOB {job_id.hex()} - Finished running {component.get_id().decode()} on "
        f"{resource_id.hex()}:{logging_component_results_str}"
    )


@lru_cache(None)
def _build_tag_filter(tags: Tuple[ResourceTag]) -> ComponentFilter:
    """
    When auto-running components, most of the time only the *most specific* components should be
    run for a resource. For example, an APK resource is also a ZIP resource; we want to always run
    the APK Unpacker on resources that are tagged as both ZIP and APK, because APK is a more
    specific tag. However, Identifiers are a special case because they have benign side-effects, so
    it is desirable to greedily run all Identifiers that could target a resource, not only the most
    specific Identifiers.

    This function constructs a filter which allows only components that target at least one of the
    given tags, but for non-identifiers the filter is even stricter so that only the most specific
    components are filtered.

    :param tags: Tags to target, from the resource that is being auto-run on

    :return: A filter which allows a component to run if (it is an Identifier, AND it targets at
    least one of the given tags) OR (it is NOT an Identifier, AND it targets one of the most
    specific given tags that are targeted by components)
    """
    tags_by_specificity = ResourceTag.sort_tags_into_tiers(tags)

    filters_prioritized_by_specificity = tuple(
        ComponentTargetFilter(*tag_specificity_level)
        for tag_specificity_level in tags_by_specificity
    )
    return ComponentOrMetaFilter(
        ComponentAndMetaFilter(
            IDENTIFIERS_FILTER,
            ComponentTargetFilter(*tags),
        ),
        ComponentAndMetaFilter(
            ComponentNotMetaFilter(
                IDENTIFIERS_FILTER,
            ),
            ComponentPrioritySelectingMetaFilter(*filters_prioritized_by_specificity),
        ),
    )


@lru_cache(None)
def _build_auto_run_filter(
    request: JobMultiComponentRequest,
) -> ComponentFilter:
    filters: List[ComponentFilter] = []
    if request.components_allowed:
        filters.append(ComponentWhitelistFilter(*request.components_allowed))

    type_filters = []
    if request.all_unpackers:
        type_filters.append(UNPACKERS_FILTER)
    if request.all_identifiers:
        type_filters.append(IDENTIFIERS_FILTER)
    if request.all_analyzers:
        type_filters.append(ANALYZERS_FILTER)
    if request.all_packers:
        type_filters.append(PACKERS_FILTER)
    filters.append(ComponentOrMetaFilter(*type_filters))
    if request.components_disallowed:
        filters.append(
            ComponentNotMetaFilter(ComponentWhitelistFilter(*request.components_disallowed))
        )
    if request.tags_ignored:
        filters.append(ComponentNotMetaFilter(ComponentTargetFilter(*request.tags_ignored)))

    return ComponentAndMetaFilter(*filters)



import asyncio
import binascii
import dataclasses
import re
from enum import Enum
import functools
import itertools
import logging
from ofrak.project.project import OfrakProject

import typing_inspect
from typing_inspect import get_args
import json
import orjson
import inspect
import os
import sys
import webbrowser
from collections import defaultdict
from typing import (
    Iterable,
    Optional,
    Dict,
    cast,
    Set,
    Tuple,
    Union,
    Type,
    Callable,
    TypeVar,
    Any,
    List,
)

from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from aiohttp.web_fileresponse import FileResponse
from dataclasses import fields

from ofrak.component.interface import ComponentInterface
from ofrak.model.component_filters import (
    ComponentOrMetaFilter,
    ComponentTypeFilter,
    ComponentTargetFilter,
    ComponentAndMetaFilter,
)
from ofrak.ofrak_context import get_current_ofrak_context
from ofrak.service.component_locator_i import ComponentFilter
from ofrak_patch_maker.toolchain.abstract import Toolchain
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range
from ofrak import (
    OFRAKContext,
    ResourceFilter,
    ResourceAttributeRangeFilter,
    ResourceAttributeValueFilter,
    ResourceSort,
    ResourceTag,
    Packer,
    Unpacker,
    Modifier,
    Analyzer,
)
from ofrak.core import Addressable, File
from ofrak.core import (
    GenericBinary,
    AddCommentModifier,
    AddCommentModifierConfig,
    DeleteCommentModifierConfig,
    DeleteCommentModifier,
    StringFindReplaceConfig,
    StringFindReplaceModifier,
)
from ofrak.model.component_model import (
    ComponentContext,
    ClientComponentContext,
    ComponentRunResult,
    ComponentConfig,
)
from ofrak.model.resource_model import (
    ResourceContext,
    ClientResourceContext,
    ResourceModel,
    ResourceAttributes,
    MutableResourceModel,
)
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak.resource import Resource
from ofrak.service.error import SerializedError
from ofrak.service.serialization.pjson import (
    SerializationServiceInterface,
    PJSONSerializationService,
)
from ofrak.gui.script_builder import ActionType, ScriptBuilder
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.core.entropy import DataSummaryAnalyzer

T = TypeVar("T")
LOGGER = logging.getLogger(__name__)


def exceptions_to_http(error_class: Type[SerializedError]):
    """
    Decorator for a server function that attempts to do some work, and
    forwards the exception, if any, to the client over HTTP.

    Usage:

    @exceptions_to_http(MyErrorClass)
    async def handle_some_request(self, request...):
        ...
    """

    def exceptions_to_http_decorator(func: Callable):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as error:
                LOGGER.exception("Exception raised in aiohttp endpoint")
                return respond_with_error(error, error_class)

        return wrapper

    return exceptions_to_http_decorator


class AiohttpOFRAKServer:
    routes = web.RouteTableDef()

    def __init__(
        self,
        serializer: SerializationServiceInterface,
        ofrak_context: OFRAKContext,
        host: str,
        port: int,
        enable_cors: bool = False,
    ):
        self._serializer = serializer
        self._app = web.Application(client_max_size=None)  # type: ignore
        self._host = host
        self._port = port
        self._ofrak_context = ofrak_context
        self.resource_context: ResourceContext = ClientResourceContext()
        self.resource_view_context: ResourceViewContext = ResourceViewContext()
        self.component_context: ComponentContext = ClientComponentContext()
        self.script_builder: ScriptBuilder = ScriptBuilder()
        self.resource_builder: Dict[str, Tuple[Resource, memoryview]] = {}
        self.projects: Optional[Set[OfrakProject]] = None
        self.projects_dir: str = "/tmp/ofrak-projects"
        self._app.add_routes(
            [
                web.post("/create_root_resource", self.create_root_resource),
                web.post("/init_chunked_root_resource", self.init_chunked_root_resource),
                web.post("/root_resource_chunk", self.root_resource_chunk),
                web.post("/create_chunked_root_resource", self.create_chunked_root_resource),
                web.get("/get_root_resources", self.get_root_resources),
                web.get("/{resource_id}/", self.get_resource),
                web.get("/{resource_id}/get_data", self.get_data),
                web.get("/{resource_id}/get_data_length", self.get_data_length),
                web.post(
                    "/batch/get_data_range_within_parent",
                    self.batch_get_range,
                ),
                web.get(
                    "/{resource_id}/get_child_data_ranges",
                    self.get_child_data_ranges,
                ),
                web.get("/{resource_id}/get_root", self.get_root_resource_from_child),
                web.post("/{resource_id}/unpack", self.unpack),
                web.post("/{resource_id}/unpack_recursively", self.unpack_recursively),
                web.post("/{resource_id}/pack", self.pack),
                web.post("/{resource_id}/pack_recursively", self.pack_recursively),
                web.post("/{resource_id}/analyze", self.analyze),
                web.post("/{resource_id}/identify", self.identify),
                web.post("/{resource_id}/data_summary", self.data_summary),
                web.get("/{resource_id}/get_parent", self.get_parent),
                web.get("/{resource_id}/get_ancestors", self.get_ancestors),
                web.get("/{resource_id}/get_descendants", self.get_descendants),
                web.post("/batch/get_children", self.batch_get_children),
                web.post("/{resource_id}/queue_patch", self.queue_patch),
                web.post("/{resource_id}/create_mapped_child", self.create_mapped_child),
                web.post("/{resource_id}/find_and_replace", self.find_and_replace),
                web.post("/{resource_id}/add_comment", self.add_comment),
                web.post("/{resource_id}/delete_comment", self.delete_comment),
                web.post("/{resource_id}/search_for_vaddr", self.search_for_vaddr),
                web.post("/{resource_id}/search_for_string", self.search_for_string),
                web.post("/{resource_id}/search_for_bytes", self.search_for_bytes),
                web.post("/{resource_id}/add_tag", self.add_tag),
                web.post(
                    "/{resource_id}/add_flush_to_disk_to_script", self.add_flush_to_disk_to_script
                ),
                web.get("/get_all_tags", self.get_all_tags),
                web.get("/{resource_id}/get_script", self.get_script),
                web.post(
                    "/{resource_id}/get_components",
                    self.get_components,
                ),
                web.get("/{resource_id}/get_config_for_component", self.get_config_for_component),
                web.post("/{resource_id}/run_component", self.run_component),
                web.post(
                    "/{resource_id}/get_tags_and_num_components", self.get_tags_and_num_components
                ),
                web.post("/{resource_id}/search_data", self.search_data),
                web.post("/create_new_project", self.create_new_project),
                web.get("/get_all_projects", self.get_all_projects),
                web.get("/get_project_by_id", self.get_project_by_id),
                web.post("/add_binary_to_project", self.add_binary_to_project),
                web.post("/add_script_to_project", self.add_script_to_project),
                web.post("/open_project", self.open_project),
                web.post("/clone_project_from_git", self.clone_project_from_git),
                web.get("/get_projects_path", self.get_projects_path),
                web.post("/set_projects_path", self.set_projects_path),
                web.post("/save_project_data", self.save_project_data),
                web.post("/delete_binary_from_project", self.delete_binary_from_project),
                web.post("/delete_script_from_project", self.delete_script_from_project),
                web.post("/reset_project", self.reset_project),
                web.get("/get_project_script", self.get_project_script),
                web.get("/", self.get_static_files),
                web.static(
                    "/",
                    os.path.join(os.path.dirname(__file__), "./public"),
                    show_index=True,
                ),
            ]
        )

        self._job_ids: Dict[str, bytes] = defaultdict(
            lambda: ofrak_context.id_service.generate_id()
        )
        self._all_tags: Dict[str, ResourceTag] = {tag.__name__: tag for tag in ResourceTag.all_tags}

        if enable_cors:
            try:
                import aiohttp_cors  # type: ignore

                # From: https://github.com/aio-libs/aiohttp-cors
                # Configure default CORS settings.
                cors = aiohttp_cors.setup(
                    self._app,
                    defaults={
                        "*": aiohttp_cors.ResourceOptions(
                            allow_credentials=True,
                            expose_headers="*",
                            allow_headers="*",
                        )
                    },
                )

                # Configure CORS on all routes.
                for route in list(self._app.router.routes()):
                    cors.add(route)
            except ImportError:
                LOGGER.warning(
                    "Unable to enable CORS. Please confirm that aiohttp_cors is installed."
                )

    async def start(self):  # pragma: no cover
        """
        Start the server then return.
        """
        self.runner = web.AppRunner(self._app)
        await self.runner.setup()
        server = web.TCPSite(self.runner, host=self._host, port=self._port)
        await server.start()

    async def stop(self):  # pragma: no cover
        """
        Stop the server.
        """
        await self.runner.server.shutdown()
        await self.runner.cleanup()

    async def run_until_cancelled(self):  # pragma: no cover
        """
        To be run after `start_server`, within an asyncio Task.
        cancel() that task to shutdown the server.
        """
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass
        finally:
            await self.runner.cleanup()

    @exceptions_to_http(SerializedError)
    async def init_chunked_root_resource(self, request: Request) -> Response:
        name = request.query.get("name")
        size_param = request.query.get("size")
        if name is None:
            raise HTTPBadRequest(reason="Missing resource name from request")
        if size_param is None:
            raise HTTPBadRequest(reason="Missing chunk size from request")
        size = int(size_param)
        root_resource: Resource = await self._ofrak_context.create_root_resource(name, b"", (File,))
        self.resource_builder[root_resource.get_id().hex()] = (
            root_resource,
            memoryview(bytearray(b"\x00" * size)),
        )
        return json_response(root_resource.get_id().hex())

    @exceptions_to_http(SerializedError)
    async def root_resource_chunk(self, request: Request) -> Response:
        id = request.query.get("id")
        start_param = request.query.get("start")
        end_param = request.query.get("end")
        if id is None:
            raise HTTPBadRequest(reason="Missing resource id from request")
        if start_param is None:
            raise HTTPBadRequest(reason="Missing chunk start from request")
        if end_param is None:
            raise HTTPBadRequest(reason="Missing chunk end from request")
        start = int(start_param)
        end = int(end_param)
        chunk_data = await request.read()
        _, data = self.resource_builder[id]
        data[start:end] = chunk_data
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def create_chunked_root_resource(self, request: Request) -> Response:
        id = request.query.get("id")
        name = request.query.get("name")
        if id is None:
            return HTTPBadRequest(reason="Missing root resource `id` from request")
        if name is None:
            return HTTPBadRequest(reason="Missing root resource `name` from request")

        try:
            root_resource, data = self.resource_builder[id]
            script_str = rf"""
            if root_resource is None:
                root_resource = await ofrak_context.create_root_resource_from_file("{name}")"""
            root_resource.queue_patch(Range(0, 0), bytearray(data))
            await root_resource.save()
            await self.script_builder.add_action(root_resource, script_str, ActionType.UNPACK)
            if request.remote is not None:
                self._job_ids[request.remote] = root_resource.get_job_id()
            await self.script_builder.commit_to_script(root_resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(root_resource)
            raise e
        self.resource_builder.pop(id)
        return json_response(self._serialize_resource(root_resource))

    @exceptions_to_http(SerializedError)
    async def create_root_resource(self, request: Request) -> Response:
        name = request.query.get("name")
        if name is None:
            return HTTPBadRequest(reason="Missing root resource `name` from request")
        resource_data = await request.read()
        script_str = rf"""
        if root_resource is None:
            root_resource = await ofrak_context.create_root_resource_from_file("{name}")"""
        try:
            root_resource = await self._ofrak_context.create_root_resource(
                name, resource_data, (File,)
            )
            await self.script_builder.add_action(root_resource, script_str, ActionType.UNPACK)
            if request.remote is not None:
                self._job_ids[request.remote] = root_resource.get_job_id()
            await self.script_builder.commit_to_script(root_resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(root_resource)
            raise e
        return json_response(self._serialize_resource(root_resource))

    @exceptions_to_http(SerializedError)
    async def get_root_resources(self, request: Request) -> Response:
        roots = await self._ofrak_context.resource_service.get_root_resources()

        return json_response(list(map(self._serialize_resource_model, roots)))

    @exceptions_to_http(SerializedError)
    async def get_resource(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_data(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        _range = self._serializer.from_pjson(
            get_query_string_as_pjson(request).get("range"), Optional[Range]
        )
        data = await resource.get_data(_range)
        return Response(body=data)

    @exceptions_to_http(SerializedError)
    async def get_data_length(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        data_len = await resource.get_data_length()
        return json_response(data_len)

    @exceptions_to_http(SerializedError)
    async def get_child_data_ranges(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        resource_service = self._ofrak_context.resource_factory._resource_service
        data_service = self._ofrak_context.resource_factory._data_service
        children = await resource_service.get_descendants_by_id(
            resource.get_id(),
            max_depth=1,
        )

        async def get_range(child):
            try:
                if child.data_id is None:
                    return
                data_range = await data_service.get_range_within_other(
                    child.data_id, resource.get_data_id()
                )
                return child.id.hex(), (data_range.start, data_range.end)
            except ValueError:
                pass

        return json_response(
            dict(filter(lambda x: x is not None, await asyncio.gather(*map(get_range, children))))
        )

    @exceptions_to_http(SerializedError)
    async def batch_get_range(self, request: Request) -> Response:
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")

        async def get_resource_range(resource_id):
            resource_model = await self._get_resource_model_by_id(
                bytes.fromhex(resource_id), job_id
            )
            if resource_model.data_id is None:
                raise ValueError(
                    "Resource does not have a data_id. Cannot get data range from a "
                    "resource with no data."
                )
            if resource_model.parent_id is None:
                data_range = Range(0, 0)
            else:
                resource_service = self._ofrak_context.resource_factory._resource_service
                data_service = self._ofrak_context.resource_factory._data_service
                parent_models = list(
                    await resource_service.get_ancestors_by_id(resource_model.id, max_count=1)
                )
                if len(parent_models) != 1:
                    raise NotFoundError(
                        f"There is no parent for resource {resource_model.id.hex()}"
                    )
                parent_model = parent_models[0]

                parent_data_id = parent_model.data_id
                if parent_data_id is None:
                    data_range = Range(0, 0)
                else:
                    try:
                        data_range = await data_service.get_range_within_other(
                            resource_model.data_id, parent_data_id
                        )
                    except ValueError:
                        data_range = Range(0, 0)
            return resource_id, [data_range.start, data_range.end]

        return json_response(
            dict(await asyncio.gather(*map(get_resource_range, await request.json())))
        )

    @exceptions_to_http(SerializedError)
    async def unpack(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.unpack()"""
        await self.script_builder.add_action(resource, script_str, ActionType.UNPACK)
        try:
            result = await resource.unpack()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def unpack_recursively(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.unpack_recursively()"""
        await self.script_builder.add_action(resource, script_str, ActionType.UNPACK)
        try:
            result = await resource.unpack_recursively()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def pack(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.pack()"""
        await self.script_builder.add_action(resource, script_str, ActionType.PACK)
        try:
            result = await resource.pack()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def pack_recursively(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.pack_recursively()"""
        await self.script_builder.add_action(resource, script_str, ActionType.PACK)
        try:
            result = await resource.pack_recursively()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def identify(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.identify()"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.identify()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def data_summary(self, request: Request) -> Response:
        resource = cast(Resource, await self._get_resource_for_request(request))
        result = await resource.run(DataSummaryAnalyzer)

        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def analyze(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        script_str = """
        await {resource}.auto_run(all_analyzers=True)"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.auto_run(all_analyzers=True)
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def get_parent(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        parent = await resource.get_parent()

        return json_response(self._serialize_resource(parent))

    @exceptions_to_http(SerializedError)
    async def get_ancestors(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        ancestors = await resource.get_ancestors()

        return json_response(self._serialize_multi_resource(ancestors))

    @exceptions_to_http(SerializedError)
    async def get_descendants(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        descendants = await resource.get_descendants()

        return json_response(self._serialize_multi_resource(descendants))

    @exceptions_to_http(SerializedError)
    async def batch_get_children(self, request: Request) -> Response:
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")

        async def get_resource_children(resource_id):
            resource = await self._get_resource_by_id(bytes.fromhex(resource_id), job_id)
            child_models = await resource._resource_service.get_descendants_by_id(
                resource._resource.id,
                max_depth=1,
            )
            serialized_children = list(map(self._serialize_resource_model, child_models))
            serialized_children.sort(key=get_child_sort_key)

            return resource_id, serialized_children

        def get_child_sort_key(child):
            attrs = dict(child.get("attributes"))
            data_attr = attrs.get("ofrak.model.resource_model.Data")
            if data_attr is not None:
                return data_attr[1]["_offset"]
            else:
                return sys.maxsize

        return json_response(
            dict(await asyncio.gather(*map(get_resource_children, await request.json())))
        )

    @exceptions_to_http(SerializedError)
    async def get_root_resource_from_child(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        parent = resource
        try:
            # Assume get_ancestors returns an ordered list with the parent first and the root last
            for parent in await resource.get_ancestors():
                pass
        except NotFoundError:
            pass

        return json_response(self._serialize_resource(parent))

    @exceptions_to_http(SerializedError)
    async def queue_patch(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        new_data = await request.read()

        start_param = request.query.get("start")
        start = int(start_param) if start_param is not None else 0
        end_param = request.query.get("end")
        end = int(end_param) if end_param is not None else (await resource.get_data_length())
        # TODO: There has to be a better way
        new_data_string = "\\x" + "\\x".join(
            [new_data.hex()[i : i + 2] for i in range(0, len(new_data.hex()), 2)]
        )
        script_str = (
            """
        {resource}.queue_patch"""
            f"""(Range({hex(start)}, {hex(end)}), b"{new_data_string}")"""
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        script_str = """
        await {resource}.save()"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            resource.queue_patch(Range(start, end), new_data)
            await resource.save()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def create_mapped_child(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        _range = self._serializer.from_pjson(await request.json(), Optional[Range])
        script_str = (
            """
        await {resource}"""
            f""".create_child(tags=(GenericBinary,), data_range={_range})
        """
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            child = await resource.create_child(tags=(GenericBinary,), data_range=_range)
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(self._serialize_resource(child))

    @exceptions_to_http(SerializedError)
    async def find_and_replace(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        config = self._serializer.from_pjson(await request.json(), StringFindReplaceConfig)
        script_str = f"""
        config = StringFindReplaceConfig(
            to_find="{config.to_find}", 
            replace_with="{config.replace_with}", 
            null_terminate={config.null_terminate}, 
            allow_overflow={config.allow_overflow}
        )"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        script_str = """
        await {resource}.run(StringFindReplaceModifier, config)"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.run(StringFindReplaceModifier, config=config)
            await self.script_builder.commit_to_script(resource)
            return json_response(await self._serialize_component_result(result))
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e

    async def add_comment(self, request: Request) -> Response:
        """
        Expected POST body is a comment in the form Tuple[Optional[Range], str] (serialized to JSON).
        """
        resource = await self._get_resource_for_request(request)
        comment = self._serializer.from_pjson(await request.json(), Tuple[Optional[Range], str])
        script_str = (
            """
        await {resource}.run"""
            f"""(AddCommentModifier, AddCommentModifierConfig({comment}))
        """
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.run(AddCommentModifier, AddCommentModifierConfig(comment))
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def delete_comment(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        comment_range = self._serializer.from_pjson(await request.json(), Optional[Range])
        script_str = (
            """
        await {resource}.run"""
            f"""(
            DeleteCommentModifier, DeleteCommentModifierConfig({comment_range})
        )"""
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.run(
                DeleteCommentModifier, DeleteCommentModifierConfig(comment_range)
            )
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def search_for_vaddr(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        vaddr_start, vaddr_end = self._serializer.from_pjson(
            await request.json(), Tuple[int, Optional[int]]
        )

        try:
            vaddr_filter: Union[ResourceAttributeRangeFilter, ResourceAttributeValueFilter]
            if vaddr_end is not None:
                vaddr_filter = ResourceAttributeRangeFilter(
                    Addressable.VirtualAddress, vaddr_start, vaddr_end
                )
            else:
                vaddr_filter = ResourceAttributeValueFilter(Addressable.VirtualAddress, vaddr_start)
            matching_resources = await resource.get_descendants(
                r_filter=ResourceFilter(attribute_filters=(vaddr_filter,)),
                r_sort=ResourceSort(Addressable.VirtualAddress),
            )

            return json_response(list(map(self._serialize_resource, matching_resources)))

        except NotFoundError:
            return json_response([])

    @exceptions_to_http(SerializedError)
    async def search_for_string(self, request: Request):
        resource = await self._get_resource_for_request(request)
        body = await request.json()
        string_query_param = body["search_query"]
        if string_query_param == "":
            return json_response(None)
        regex = body["regex"]
        case_ignore = body["caseIgnore"]
        if not isinstance(string_query_param, str):
            raise ValueError("Invalid search query.")
        string_query: Union[bytes, re.Pattern[bytes]] = string_query_param.encode()
        try:
            if case_ignore:
                if not regex:
                    string_query = re.compile(re.escape(string_query_param.encode()), re.IGNORECASE)
                else:
                    string_query = re.compile(string_query_param.encode(), re.IGNORECASE)
            elif regex:
                string_query = re.compile(string_query_param.encode())
        except re.error:
            logging.exception("Bad regex expression in search")
        offsets = await resource.search_data(string_query)
        found_resources = []
        if len(offsets) > 0:
            found_resources.append(resource.get_id().hex())
        for child in await resource.get_descendants():
            if child.get_data_id() is None:
                continue
            offsets = await child.search_data(string_query)
            if len(offsets) > 0:
                found_resources.append(child.get_id().hex())
                for ancestor in await child.get_ancestors():
                    found_resources.append(ancestor.get_id().hex())

        return json_response(found_resources)

    @exceptions_to_http(SerializedError)
    async def search_for_bytes(self, request: Request):
        resource = await self._get_resource_for_request(request)
        body = await request.json()
        bytes_query_request = body["search_query"]
        if bytes_query_request == "":
            return json_response(None)
        bytes_query = bytes.fromhex(re.sub(r"[^0-9a-fA-F]+", "", bytes_query_request))
        offsets = await resource.search_data(bytes_query)
        found_resources = []
        if len(offsets) > 0:
            found_resources.append(resource.get_id().hex())
        for child in await resource.get_descendants():
            offsets = await child.search_data(bytes_query)
            if len(offsets) > 0:
                found_resources.append(child.get_id().hex())
                for ancestor in await child.get_ancestors():
                    found_resources.append(ancestor.get_id().hex())

        return json_response(found_resources)

    @exceptions_to_http(SerializedError)
    async def add_tag(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        tag = self._serializer.from_pjson(await request.json(), ResourceTag)
        script_str = (
            """
        {resource}.add_tag"""
            f"""({tag.__name__})"""
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        script_str = """
        await {resource}.save()"""
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            resource.add_tag(tag)
            await resource.save()
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_all_tags(self, request: Request) -> Response:
        return json_response(
            self._serializer.to_pjson(set(self._all_tags.values()), Set[ResourceTag])
        )

    @exceptions_to_http(SerializedError)
    async def add_flush_to_disk_to_script(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        output_file_name = self._serializer.from_pjson(await request.json(), str)
        # Use FilesystemEntry name as filename if available, otherwise generate random filename
        if not output_file_name:
            output_file_name = self._ofrak_context.id_service.generate_id().hex()
        script_str = (
            """
        await {resource}"""
            f""".flush_to_disk("{output_file_name}")"""
        )
        await self.script_builder.add_action(resource, script_str, ActionType.UNDEF)
        await self.script_builder.commit_to_script(resource)
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_script(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        return json_response(await self.script_builder.get_script(resource))

    @exceptions_to_http(SerializedError)
    async def get_components(self, request: Request) -> Response:
        resource: Resource = await self._get_resource_for_request(request)
        options = await request.json()
        show_all_components = options["show_all_components"]
        target_filter = options["target_filter"]
        incl_analyzers = options["analyzers"]
        incl_modifiers = options["modifiers"]
        incl_packers = options["packers"]
        incl_unpackers = options["unpackers"]
        components = self._get_specific_components(
            resource,
            show_all_components,
            target_filter,
            incl_analyzers,
            incl_modifiers,
            incl_packers,
            incl_unpackers,
        )
        return json_response(self._serializer.to_pjson(components, Set[str]))

    @exceptions_to_http(SerializedError)
    async def get_config_for_component(self, request: Request) -> Response:
        component_string = request.query.get("component")
        if component_string is not None:
            component = self._ofrak_context.component_locator.get_by_id(
                component_string.encode("ascii")
            )
            config = self._get_config_for_component(type(component))
        else:
            return json_response([])
        if (
            not config == inspect._empty
            and not type(config) == type(None)
            and not typing_inspect.is_optional_type(config)
        ):
            _fields = []
            for field in fields(config):
                field.type = self._modify_by_case(field.type)
                if isinstance(field.default, dataclasses._MISSING_TYPE):
                    field.default = None
                _fields.append(
                    {
                        "name": field.name,
                        "type": self._convert_to_class_name_str(field.type),
                        "args": self._construct_arg_response(field.type),
                        "fields": self._construct_field_response(field.type),
                        "enum": self._construct_enum_response(field.type),
                        "default": field.default
                        if not isinstance(field.default, dataclasses._MISSING_TYPE)
                        else None,
                    }
                )
            return json_response(
                {
                    "name": config.__name__,
                    "type": self._convert_to_class_name_str(config),
                    "args": self._construct_arg_response(self._convert_to_class_name_str(config)),
                    "enum": self._construct_enum_response(config),
                    "fields": _fields,
                }
            )
        else:
            return json_response([])

    @exceptions_to_http(SerializedError)
    async def run_component(self, request: Request) -> Response:
        resource: Resource = await self._get_resource_for_request(request)
        component_string = request.query.get("component")
        if component_string is not None:
            component = type(
                self._ofrak_context.component_locator.get_by_id(component_string.encode("ascii"))
            )
            config_type = self._get_config_for_component(component)
        else:
            return json_response([])
        if config_type == inspect._empty:
            config = None
        else:
            config = self._serializer.from_pjson(await request.json(), config_type)

        config_str = str(config).replace("{", "{{").replace("}", "}}")
        script_str = (
            """
        await {resource}"""
            f""".run({request.query.get("component")}, {config_str})"""
        )
        await self.script_builder.add_action(resource, script_str, ActionType.MOD)
        try:
            result = await resource.run(component, config)
            await self.script_builder.commit_to_script(resource)
        except Exception as e:
            await self.script_builder.clear_script_queue(resource)
            raise e
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def get_static_files(self, request: Request) -> FileResponse:
        return FileResponse(os.path.join(os.path.dirname(__file__), "./public/index.html"))

    @exceptions_to_http(SerializedError)
    async def get_tags_and_num_components(self, request: Request):
        resource = await self._get_resource_for_request(request)
        options = await request.json()
        only_target = options["target"]
        incl_analyzers = options["analyzers"]
        incl_modifiers = options["modifiers"]
        incl_packers = options["packers"]
        incl_unpackers = options["unpackers"]
        all_resource_tags: Set[Tuple[str, int]] = set()
        for specific_tag in resource.get_most_specific_tags():
            for tag in specific_tag.tag_classes():
                components = self._get_specific_components(
                    resource,
                    only_target,
                    tag.__qualname__,
                    incl_analyzers,
                    incl_modifiers,
                    incl_packers,
                    incl_unpackers,
                )
                all_resource_tags.add((tag.__qualname__, len(components)))
        all_resource_tags_l: List[Tuple[str, int]] = list(all_resource_tags)
        for resource_tag in all_resource_tags_l:
            if "object" in resource_tag:
                all_resource_tags_l.remove(resource_tag)
        return json_response(all_resource_tags_l)

    @exceptions_to_http(SerializedError)
    async def search_data(self, request: Request) -> Response:
        resource: Resource = await self._get_resource_for_request(request)
        body = await request.json()
        mode = body.get("searchType")
        regex = body.get("regex")
        case_ignore = body.get("caseIgnore")
        raw_query = body.get("search_query")
        if mode is None:
            mode = "String"

        if mode == "String":
            query = raw_query.encode("utf-8")

            if regex and case_ignore:
                query = re.compile(query, re.IGNORECASE)
            elif regex:
                query = re.compile(query)
            elif case_ignore:
                query = re.compile(re.escape(query), re.IGNORECASE)

        elif mode == "Bytes":
            if regex:
                raise NotImplementedError("regex for bytes not yet supported")
            query = binascii.unhexlify(raw_query.replace(" ", ""))

        else:
            raise ValueError(f"Invalid query mode {mode}")
        results = await resource.search_data(query)
        if isinstance(query, bytes):
            results = [(offset, len(query)) for offset in results]
        else:
            # final search query was regex pattern, matches were also returned
            results = [(offset, len(match)) for offset, match in results]

        return json_response(results)

    @exceptions_to_http(SerializedError)
    async def create_new_project(self, request: Request) -> Response:
        if self.projects is None:
            self.projects = self._slurp_projects_from_dir()
        body = await request.json()
        name = body.get("name")
        project = OfrakProject.create(name, os.path.join(self.projects_dir, name))
        self.projects.add(project)

        return json_response({"id": project.session_id.hex()})

    @exceptions_to_http(SerializedError)
    async def clone_project_from_git(self, request: Request) -> Response:
        if self.projects is None:
            self.projects = self._slurp_projects_from_dir()

        def recurse_path_collisions(path: str, count: int) -> str:
            if count == 0:
                incr_path = path
            else:
                incr_path = f"{path}_{count}"
            if os.path.exists(incr_path):
                count += 1
                return recurse_path_collisions(path, count)
            else:
                return incr_path

        body = await request.json()
        url = body.get("url")
        path = recurse_path_collisions(
            os.path.join(self.projects_dir, url.split(":")[-1].split("/")[-1]), 0
        )
        project = OfrakProject.clone_from_git(url, path)
        self.projects.add(project)
        return json_response({"id": project.session_id.hex()})

    @exceptions_to_http(SerializedError)
    async def get_project_by_id(self, request: Request) -> Response:
        id = request.query.get("id")
        project = self._get_project_by_id(id)
        return json_response(project.get_current_metadata())

    @exceptions_to_http(SerializedError)
    async def get_all_projects(self, request: Request) -> Response:
        if self.projects is None:
            self.projects = self._slurp_projects_from_dir()
        return json_response([project.get_current_metadata() for project in self.projects])

    @exceptions_to_http(SerializedError)
    async def reset_project(self, request: Request) -> Response:
        body = await request.json()
        id = body["id"]
        project = self._get_project_by_id(id)
        project.reset_project()
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def add_binary_to_project(self, request: Request) -> Response:
        id = request.query.get("id")
        name_query = request.query.get("name")
        if name_query is not None:
            name = name_query
        data = await request.read()
        project = self._get_project_by_id(id)
        project.add_binary(name, data)
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def add_script_to_project(self, request: Request) -> Response:
        id = request.query.get("id")
        name_query = request.query.get("name")
        if name_query is not None:
            name = name_query
        data = await request.read()
        project = self._get_project_by_id(id)
        project.add_script(name, data.decode())
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def open_project(self, request: Request) -> Response:
        body = await request.json()
        id = body["id"]
        binary = body["binary"]
        script = body["script"]
        if request.remote is not None:
            resource_id = request.remote
        else:
            raise AttributeError("No resource ID provided")
        project = self._get_project_by_id(id)
        resource = await project.init_project_binary(
            binary, self._ofrak_context, script_name=script
        )
        self._job_ids[resource_id] = resource.get_job_id()
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_projects_path(self, request: Request) -> Response:
        return json_response(self.projects_dir)

    @exceptions_to_http(SerializedError)
    async def set_projects_path(self, request: Request) -> Response:
        body = await request.json()
        new_path = body["path"]
        if not os.path.exists(new_path):
            os.mkdir(new_path)
        self.projects_dir = new_path
        self.projects = self._slurp_projects_from_dir()
        return json_response(self.projects_dir)

    @exceptions_to_http(SerializedError)
    async def save_project_data(self, request: Request) -> Response:
        body = await request.json()
        session_id = body["session_id"]
        project = self._get_project_by_id(session_id)
        for binary_name, binary_metadata in body["binaries"].items():
            project.binaries[binary_name].init_script = binary_metadata["init_script"]
            project.binaries[binary_name].associated_scripts = binary_metadata["associated_scripts"]
        project.write_metadata_to_disk()
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def delete_binary_from_project(self, request: Request) -> Response:
        body = await request.json()
        id = body["id"]
        binary_name = body["binary"]
        project = self._get_project_by_id(id)
        project.delete_binary(binary_name)
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def delete_script_from_project(self, request: Request) -> Response:
        body = await request.json()
        id = body["id"]
        script_name = body["script"]
        project = self._get_project_by_id(id)
        project.delete_script(script_name)
        return json_response([])

    @exceptions_to_http(SerializedError)
    async def get_project_script(self, request: Request) -> Response:
        project_id = request.query.get("project")
        script_name_query = request.query.get("script")
        if script_name_query is not None:
            script_name = script_name_query
        project = self._get_project_by_id(project_id)
        script_body = project.get_script_body(script_name)

        return Response(text=script_body)

    def _slurp_projects_from_dir(self) -> Set:
        projects = set()
        if not os.path.exists(self.projects_dir):
            os.makedirs(self.projects_dir)
        for dir in os.listdir(self.projects_dir):
            try:
                project = OfrakProject.init_from_path(os.path.join(self.projects_dir, dir))
                projects.add(project)
            except:
                pass
        return projects

    def _get_project_by_id(self, id) -> OfrakProject:
        if self.projects is None:
            self.projects = self._slurp_projects_from_dir()
        result = [project for project in self.projects if project.session_id.hex() == id]
        if len(result) > 1:
            raise AttributeError("Project ID Collision")
        if len(result) == 0:
            raise ValueError(f"Project with ID {id} not found")
        return result[0]

    def _construct_field_response(self, obj):
        if dataclasses.is_dataclass(obj):
            res = []
            for field in fields(obj):
                if field.init:
                    field.type = self._modify_by_case(field.type)
                    res.append(
                        {
                            "name": field.name,
                            "type": self._convert_to_class_name_str(field.type),
                            "args": self._construct_arg_response(field.type),
                            "fields": self._construct_field_response(field.type),
                            "enum": self._construct_enum_response(field.type),
                            "default": field.default
                            if not isinstance(field.default, dataclasses._MISSING_TYPE)
                            else None,
                        }
                    )
            return res
        else:
            return None

    def _construct_arg_response(self, obj):
        args = get_args(obj)
        if len(args) != 0:
            res = []
            for arg in args:
                arg = self._modify_by_case(arg)
                res.append(
                    {
                        "name": None,
                        "type": self._convert_to_class_name_str(arg),
                        "args": self._construct_arg_response(arg),
                        "fields": self._construct_field_response(arg),
                        "enum": self._construct_enum_response(arg),
                        "default": None,
                    }
                )
            return res
        else:
            return None

    def _modify_by_case(self, obj):
        args = get_args(obj)
        if self._has_elipsis(obj):
            if len(args) == 2:
                other_arg = [arg for arg in args if not isinstance(arg, type(...))][0]
                obj = List[other_arg]
            else:
                raise AttributeError("Unexpected type format with elipsis")
        return obj

    def _construct_enum_response(self, obj):
        if obj == Type[Toolchain]:
            return {
                tc.__name__: f"{tc.__module__}.{tc.__qualname__}"
                for tc in Toolchain.toolchain_implementations
                if not inspect.isabstract(tc)
            }
        if not inspect.isclass(obj):
            return None
        elif not issubclass(obj, Enum):
            return None
        else:
            return {name: value.value for name, value in obj.__members__.items()}

    def _has_elipsis(self, obj):
        return any([isinstance(arg, type(...)) for arg in get_args(obj)])

    def _convert_to_class_name_str(self, obj: Any):
        if hasattr(obj, "__qualname__") and hasattr(obj, "__module__"):
            return f"{obj.__module__}.{obj.__qualname__}"
        else:
            if obj in {bool, str, bytes, int}:
                return f"builtins.{obj.__name__}"
            elif typing_inspect.is_optional_type(obj):
                return "typing.Optional"
            elif typing_inspect.is_union_type(obj):
                return "typing.Union"
            elif obj is Range:
                return "ofrak_type.range.Range"
            elif hasattr(obj, "__origin__"):
                origin = obj.__origin__
                if origin is list:
                    return "typing.List"
                elif origin is Iterable.__origin__:  # type: ignore
                    return "typing.Iterable"
                elif origin is tuple:
                    return "typing.Tuple"
                elif origin is dict:
                    return "typing.Dict"
            else:
                return repr(obj).split("[")[0]

    async def _get_resource_by_id(self, resource_id: bytes, job_id: bytes) -> Resource:
        resource = await self._ofrak_context.resource_factory.create(
            job_id,
            resource_id,
            self.resource_context,
            self.resource_view_context,
            self.component_context,
        )
        return resource

    def _get_specific_components(
        self,
        resource: Resource,
        show_all_components: bool,
        target_filter: Optional[str],
        incl_analyzers: bool,
        incl_modifiers: bool,
        incl_packers: bool,
        incl_unpackers: bool,
    ) -> List[str]:
        selected_components = []
        tags = resource.get_tags()
        if show_all_components and len(set(tags)) == 0:
            return []

        requested_components = [incl_analyzers, incl_modifiers, incl_packers, incl_unpackers]
        categories: Tuple[Type[ComponentInterface], ...] = (Analyzer, Modifier, Packer, Unpacker)
        if any(requested_components):
            categories = tuple(itertools.compress(categories, requested_components))

        component_filters: List[ComponentFilter] = [
            ComponentOrMetaFilter(*(ComponentTypeFilter(cat) for cat in categories)),
        ]
        if not show_all_components:
            component_filters.append(ComponentTargetFilter(*tags))
            if target_filter is not None:
                component_filters.append(ComponentTargetFilter(self._all_tags[target_filter]))

        for component in self._ofrak_context.component_locator.get_components_matching_filter(
            ComponentAndMetaFilter(*component_filters)
        ):
            if type(component).__name__ != component.get_id().decode("ascii"):
                # TODO: The server lookups for these components won't work yet
                continue
            if type(component).__name__ == "AngrAnalyzer":
                # TODO: The config for this includes some angr types and can't be serialized
                continue
            selected_components.append(type(component).__name__)

        return selected_components

    def _get_config_for_component(
        self, component: Type[ComponentInterface]
    ) -> Type[ComponentConfig]:
        if issubclass(component, Packer):
            config = inspect.signature(component.pack).parameters["config"].annotation
        elif issubclass(component, Unpacker):
            config = inspect.signature(component.unpack).parameters["config"].annotation
        elif issubclass(component, Modifier):
            config = inspect.signature(component.modify).parameters["config"].annotation
        elif issubclass(component, Analyzer):
            config = inspect.signature(component.analyze).parameters["config"].annotation
        else:
            raise ValueError("{component} can not be run from the web API.")
        if hasattr(config, "_name"):
            if config._name == "Optional":
                config = [conf for conf in get_args(config) if conf is not None][0]
        return config

    async def _get_resource_model_by_id(
        self, resource_id: bytes, job_id: bytes
    ) -> Optional[Union[ResourceModel, MutableResourceModel]]:
        resource_m: Optional[Union[ResourceModel, MutableResourceModel]] = None
        resource_m = self.resource_context.resource_models.get(resource_id)
        if resource_m is None:
            resource_m = await self._ofrak_context.resource_factory._resource_service.get_by_id(
                resource_id
            )
        return resource_m

    async def _serialize_component_result(self, result: ComponentRunResult) -> PJSONType:
        async def get_and_serialize(resource_id) -> PJSONType:
            resource_model = await self._ofrak_context.resource_service.get_by_id(resource_id)
            return self._serialize_resource_model(resource_model)

        serialized_result = {
            "created": await asyncio.gather(*map(get_and_serialize, result.resources_created)),
            "modified": await asyncio.gather(
                *map(
                    get_and_serialize,
                    result.resources_modified.difference(result.resources_created).difference(
                        result.resources_deleted
                    ),
                )
            ),
            "deleted": self._serializer.to_pjson(result.resources_deleted, Set[bytes]),
        }
        return serialized_result

    async def _get_resource_for_request(self, request: Request) -> Resource:
        resource_id = pluck_id(request, "resource_id")
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")
        return await self._get_resource_by_id(resource_id, job_id)

    def _serialize_resource_model(self, resource_model: ResourceModel) -> PJSONType:
        """
        Serialize the resource model, stripped of information irrelevant to the frontend.
        """
        result = {
            "id": resource_model.id.hex(),
            "data_id": resource_model.data_id.hex() if resource_model.data_id else None,
            "parent_id": resource_model.parent_id.hex() if resource_model.parent_id else None,
            "tags": [tag.__module__ + "." + tag.__qualname__ for tag in resource_model.tags],
            "attributes": self._serializer.to_pjson(
                resource_model.attributes, Dict[Type[ResourceAttributes], ResourceAttributes]
            ),
            "caption": resource_model.caption,
        }
        return result

    def _serialize_resource(self, resource: Resource) -> PJSONType:
        """
        Serialize the resource as a serialized model, stripped of information irrelevant to the
        frontend.
        """
        return self._serialize_resource_model(resource.get_model())

    def _serialize_multi_resource(self, resources: Iterable[Resource]) -> PJSONType:
        """
        Serialize the resources as serialized models, stripped of information irrelevant to the
        frontend.
        """
        return list(map(self._serialize_resource, resources))


async def start_server(
    ofrak_context: OFRAKContext,
    host: str,
    port: int,
    enable_cors: bool = False,
) -> AiohttpOFRAKServer:  # pragma: no cover
    # Force using the correct PJSON serialization with the expected structure. Otherwise the
    # dependency injector may accidentally use the Stashed PJSON serialization service,
    # which returns PJSON that has a different, problematic structure.
    ofrak_context.injector.bind_factory(PJSONSerializationService)

    ofrak_context.injector.bind_factory(
        AiohttpOFRAKServer,
        ofrak_context=ofrak_context,
        host=host,
        port=port,
        enable_cors=enable_cors,
    )
    server = await ofrak_context.injector.get_instance(AiohttpOFRAKServer)
    await server.start()

    return server


def respond_with_error(error: Exception, error_cls: Type[SerializedError]) -> Response:
    if isinstance(error, error_cls):
        text = error.to_json()
    else:
        text = json.dumps(error_cls.to_dict(error))
    response = Response(text=text, status=500)
    return response


def pluck_id(request: Request, get_parameter_name: str) -> bytes:
    return bytes.fromhex(request.match_info[get_parameter_name])


def get_query_string_as_pjson(request: Request) -> Dict[str, PJSONType]:
    """
    URL-encoded GET parameters are all strings. For example, None is encoded as 'None',
    or 1 as '1', which isn't valid PJSON. We fix this by applying `json.loads` on each parameter.
    """
    return {key: json.loads(value) for key, value in request.query.items()}


async def open_gui(
    host: str,
    port: int,
    focus_resource: Optional[Resource] = None,
    ofrak_context: Optional[OFRAKContext] = None,
    open_in_browser: bool = True,
    enable_cors: bool = False,
) -> AiohttpOFRAKServer:  # pragma: no cover
    if ofrak_context is None:
        ofrak_context = get_current_ofrak_context()

    server = await start_server(ofrak_context, host, port, enable_cors)

    if focus_resource is None:
        url = f"http://{server._host}:{server._port}/"
    else:
        url = f"http://{server._host}:{server._port}/#{focus_resource.get_id().hex()}"
    print(f"GUI is being served on {url}")

    if open_in_browser:
        webbrowser.open(url)

    return server


def json_response(
    data: Any = None,
    *,
    text: Optional[str] = None,
    body: Optional[bytes] = None,
    status: int = 200,
    reason: Optional[str] = None,
    headers=None,
    content_type: str = "application/json",
    dumps=orjson.dumps,
) -> Response:
    if data is not None:
        if text or body:
            raise ValueError("only one of data, text, or body should be specified")
        else:
            body = dumps(data)
    return Response(
        text=text,
        body=body,
        status=status,
        reason=reason,
        headers=headers,
        content_type=content_type,
    )



from ofrak.component.identifier import Identifier
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.patch_maker.linkable_symbol import LinkableSymbol
from ofrak.resource import Resource


class LinkableSymbolIdentifier(Identifier):
    """
    Component to identify Complex Blocks as Linkable Symbols. If this component is discovered,
    it will tag all [ComplexBlock][ofrak.core.complex_block.ComplexBlock]s as LinkableSymbols.
    """

    id = b"LinkableSymbolIdentifier"
    targets = (ComplexBlock,)

    async def identify(self, resource: Resource, config=None):
        resource.add_tag(LinkableSymbol)



import functools
import importlib
import importlib.util
import logging
import os.path
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentParser, _SubParsersAction, Namespace
from inspect import isabstract
from types import ModuleType
from typing import Dict, Iterable, List, Optional, Sequence, Type, Set

from importlib_metadata import entry_points

from ofrak.component.interface import ComponentInterface
from ofrak.model.component_model import ComponentExternalTool
from ofrak.ofrak_context import OFRAKContext, OFRAK
from synthol.injector import DependencyInjector

BACKEND_PACKAGES: Dict[Optional[str], Set[str]] = {
    "angr": {"ofrak_angr", "ofrak_capstone"},
    "binary-ninja": {"ofrak_binary_ninja", "ofrak_capstone"},
    "ghidra": {"ofrak_ghidra"},
    None: set(),
}


class OFRAKEnvironment:
    def __init__(self):
        self._ofrak_packages: Optional[Dict[str, ModuleType]] = None
        self._ofrak_components: Optional[Dict[str, Type[ComponentInterface]]] = None
        self._ofrak_package_components: Optional[
            Dict[ModuleType, List[Type[ComponentInterface]]]
        ] = None
        self._ofrak_component_dependencies: Optional[
            Dict[Type[ComponentInterface], List[ComponentExternalTool]]
        ] = None

        _OFRAK_PACKAGES: Optional[List[ModuleType]] = None
        _OFRAK_COMPONENTS: Optional[Dict[ModuleType, List[Type[ComponentInterface]]]] = None
        _OFRAK_DEPENDENCIES: Optional[
            Dict[Type[ComponentInterface], List[ComponentExternalTool]]
        ] = None

        _OFRAK_PACKAGES_BY_NAME: Optional[Dict[str, ModuleType]] = None
        _OFRAK_COMPONENTS_BY_NAME: Optional[Dict[str, Type[ComponentInterface]]] = None

    @property
    def packages(self) -> Dict[str, ModuleType]:
        if self._ofrak_packages is None:
            ofrak_eps = entry_points(group="ofrak.packages")
            import ofrak

            installed_ofrak_pkgs = [ofrak] + [ofrak_pkg.load() for ofrak_pkg in ofrak_eps]
            self._ofrak_packages = {pkg.__name__: pkg for pkg in installed_ofrak_pkgs}
        return self._ofrak_packages

    @property
    def components(self) -> Dict[str, Type[ComponentInterface]]:
        if self._ofrak_components is None:
            self._ofrak_components = {}
            for components in self.components_by_package.values():
                for component in components:
                    self._ofrak_components[component.__name__] = component

        return self._ofrak_components

    @property
    def components_by_package(self) -> Dict[ModuleType, List[Type[ComponentInterface]]]:
        if self._ofrak_package_components is None:
            self._ofrak_package_components = {}
            prev_packages: List[ModuleType] = []
            next_packages = list(self.packages.values())
            while next_packages:
                package = next_packages.pop(0)
                other_packages = prev_packages + next_packages
                injector = DependencyInjector()
                # Make sure that if a package has imports from another package, injector does not
                # follow this reference and discover code from the other package
                injector.discover(package, blacklisted_modules=other_packages)
                prev_packages.append(package)

                self._ofrak_package_components[package] = []

                for provider in injector._providers[ComponentInterface]:
                    if isabstract(provider._factory):
                        continue
                    component_type = provider._factory
                    self._ofrak_package_components[package].append(component_type)

        return self._ofrak_package_components

    @property
    def dependencies_by_component(
        self,
    ) -> Dict[Type[ComponentInterface], List[ComponentExternalTool]]:
        if self._ofrak_component_dependencies is None:
            ofrak_components = self.components.values()
            self._ofrak_component_dependencies = {
                c: list(c.external_dependencies) for c in ofrak_components  # type: ignore
            }
        return self._ofrak_component_dependencies


class OfrakCommand(ABC):
    @abstractmethod
    def create_parser(self, ofrak_subparser: _SubParsersAction):
        raise NotImplementedError()

    @abstractmethod
    def run(self, ofrak_env: OFRAKEnvironment, args: Namespace):
        raise NotImplementedError()


class OfrakCommandRunsScript(OfrakCommand, ABC):
    """
    An OFRAK CLI command that needs to set up a full OFRAK Context and use it to run an OFRAK
    script. Has additional scaffolding to make this process easy.
    """

    @staticmethod
    def add_ofrak_arguments(command_subparser):
        command_subparser.add_argument(
            "--logging-level",
            "-l",
            help="Minimum level of messages to print",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            default=OFRAK.DEFAULT_LOG_LEVEL,
        )
        command_subparser.add_argument(
            "--exclude-components-missing-dependencies",
            "-x",
            help="When initializing OFRAK, check each component's dependency and do not use any "
            "components missing some dependencies",
            action="store_true",
        )
        command_subparser.add_argument(
            "-b",
            "--backend",
            action="store",
            help="Set GUI server backend by importing additional OFRAK packages. Equivalent to "
            "importing these packages with `-i`. For example, `-b angr` is: `-i ofrak_angr "
            "-i ofrak_capstone`",
            choices=BACKEND_PACKAGES.keys(),
            default=None,
        )
        command_subparser.add_argument(
            "-i",
            "--import",
            help="Import additional OFRAK Python packages, where additional Components, Tags, etc. "
            "may be defined. Can be either the name of an installed module or a local path to "
            "a Python file.",
            required=False,
            action="append",
            default=[],
            dest="imports",  # object member can't be called "import"
        )

    def run(self, ofrak_env: OFRAKEnvironment, args: Namespace):
        if type(args.logging_level) is int:
            logging_level = args.logging_level
        else:
            logging_level = getattr(logging, args.logging_level.upper())
        ofrak = OFRAK(
            logging_level=logging_level,
            exclude_components_missing_dependencies=args.exclude_components_missing_dependencies,
        )

        ofrak_pkgs = set(args.imports)
        ofrak_pkgs.update(BACKEND_PACKAGES[args.backend])

        if not any(pkgs.issubset(ofrak_pkgs) for pkgs in BACKEND_PACKAGES.values() if pkgs):
            logging.warning("No disassembler backend specified, so no disassembly will be possible")

        for ofrak_pkg_name in ofrak_pkgs:
            if os.path.exists(ofrak_pkg_name) and os.path.isfile(os.path.abspath(ofrak_pkg_name)):
                ofrak_pkg = self._import_via_path(ofrak_pkg_name)
            else:
                ofrak_pkg = importlib.import_module(ofrak_pkg_name)
            ofrak.discover(ofrak_pkg)

        ofrak.run(self.ofrak_func, args)

    def _import_via_path(self, path: str):
        ofrak_file_path = os.path.abspath(path)
        ofrak_file_dir = os.path.dirname(ofrak_file_path)
        ofrak_file_name = os.path.basename(ofrak_file_path)
        if os.path.isdir(ofrak_file_path) or not ofrak_file_name.endswith(".py"):
            raise ImportError(f"Cannot import {path} as it does not appear to be a Python file!")
        ofrak_file_name = ofrak_file_name[:-3]

        sys.path.append(ofrak_file_dir)
        pkg = importlib.import_module(ofrak_file_name)
        return pkg

    @abstractmethod
    async def ofrak_func(self, ofrak_context: OFRAKContext, args: Namespace):
        raise NotImplementedError()


class OFRAKCommandLineInterface:
    def __init__(
        self,
        subcommands: Iterable[OfrakCommand],
        ofrak_env: OFRAKEnvironment = OFRAKEnvironment(),
    ):
        self.ofrak_parser = ArgumentParser(prog="ofrak")
        ofrak_subparsers = self.ofrak_parser.add_subparsers(
            help="Command line utilities to use or configure OFRAK"
        )

        for ofrak_subcommand in subcommands:
            subparser = ofrak_subcommand.create_parser(ofrak_subparsers)
            subparser.set_defaults(run=functools.partial(ofrak_subcommand.run, ofrak_env))

    def parse_and_run(self, args: Sequence[str]):
        parsed = self.ofrak_parser.parse_args(args)
        if not hasattr(parsed, "run"):
            self.ofrak_parser.print_help()
            sys.exit(1)
        parsed.run(parsed)
